/*
 * SIMD functions.  These replace the functions in asm-i386/string.h
 * whenever it makes sense.  These also un-inline those functions.
 *
 * Copyright 1999, Doug Ledford <dledford@redhat.com>
 *
 * These functions are simple and trivial, consider them to be
 * public domain
 */

#include <linux/string.h>
#include <linux/sched.h>
#include <asm/processor.h>
#include <asm/uaccess.h>

/*
 * We declare our accelerator functions here since this is the only place
 * that needs the declarations which makes a header file a pain to deal
 * with
 */
extern void * kni_memcpy(void *, const void *, size_t);
extern void * kni_memset(void *, char, size_t);
extern unsigned long kni_copy_to_user(void *, const void *, unsigned long);
extern unsigned long kni_copy_from_user(void *, const void *, unsigned long);
extern unsigned long __kni_copy_to_user_nocheck(void *, const void *, unsigned long);
extern unsigned long __kni_copy_from_user_nocheck(void *, const void *, unsigned long);

static void * best_memcpy_final(void *, const void *, size_t);
static void * best_memset_final(void *, char, size_t);
static unsigned long best_copy_to_user_final(void *, const void *, unsigned long);
static unsigned long best_copy_from_user_final(void *, const void *, unsigned long);
static unsigned long __best_copy_to_user_final(void *, const void *, unsigned long);
static unsigned long __best_copy_from_user_final(void *, const void *, unsigned long);

void * best_memcpy(void * to, const void * from, size_t n)
{
	int BAR = (int)__builtin_return_address(0);
	int *caller = (int *)BAR - 1;
	if(boot_cpu_data.enable_fixups) {
	    if ( (boot_cpu_data.mmu_cr4_features & X86_CR4_OSFXSR) &&
		 (boot_cpu_data.x86_capability & X86_FEATURE_XMM) ) {
			*caller = (int)kni_memcpy - BAR;
			return(kni_memcpy(to, from, n));
		} else {
			*caller = (int)best_memcpy_final - BAR;
			return(__memcpy(to, from, n));
		}
	} else {
		return(__memcpy(to, from, n));
	}
}

static void * best_memcpy_final(void * to, const void * from, size_t n)
{
	return(__memcpy(to, from, n));
}

void * best_memset(void * s, char c, size_t count)
{
	int BAR = (int)__builtin_return_address(0);
	int *caller = (int *)BAR - 1;
	if(boot_cpu_data.enable_fixups) {
	    if ( (boot_cpu_data.mmu_cr4_features & X86_CR4_OSFXSR) &&
		 (boot_cpu_data.x86_capability & X86_FEATURE_XMM) ) {
			*caller = (int)kni_memset - BAR;
			return(kni_memset(s, c, count));
		} else {
			*caller = (int)best_memset_final - BAR;
			return(__memset_generic(s, c, count));
		}
	} else {
		return(__memset_generic(s, c, count));
	}
}

static void * best_memset_final(void * s, char c, size_t count)
{
	return(__memset_generic(s, c, count));
}

unsigned long
best_copy_to_user(void *to, const void *from, unsigned long n)
{
	int BAR = (int)__builtin_return_address(0);
	int *caller = (int *)BAR - 1;
	if(boot_cpu_data.enable_fixups) {
	    if ( (boot_cpu_data.mmu_cr4_features & X86_CR4_OSFXSR) &&
		 (boot_cpu_data.x86_capability & X86_FEATURE_XMM) ) {
			*caller = (int)kni_copy_to_user - BAR;
			return(kni_copy_to_user(to, from, n));
		} else {
			*caller = (int)best_copy_to_user_final - BAR;
			return(best_copy_to_user_final(to, from, n));
		}
	} else {
		if (access_ok(VERIFY_WRITE, to, n)) {
			__copy_user(to,from,n);
		}
		return n;
	}
}

static unsigned long
best_copy_to_user_final(void *to, const void *from, unsigned long n)
{
	if (access_ok(VERIFY_WRITE, to, n)) {
		__copy_user(to,from,n);
	}
	return n;
}

unsigned long
best_copy_from_user(void *to, const void *from, unsigned long n)
{
	int BAR = (int)__builtin_return_address(0);
	int *caller = (int *)BAR - 1;
	if(boot_cpu_data.enable_fixups) {
	    if ( (boot_cpu_data.mmu_cr4_features & X86_CR4_OSFXSR) &&
		 (boot_cpu_data.x86_capability & X86_FEATURE_XMM) ) {
			*caller = (int)kni_copy_from_user - BAR;
			return(kni_copy_from_user(to, from, n));
		} else {
			*caller = (int)best_copy_from_user_final - BAR;
			return(best_copy_from_user_final(to, from, n));
		}
	} else {
		if (access_ok(VERIFY_READ, from, n)) {
			__copy_user_zeroing(to,from,n);
		}
		return n;
	}
}

static unsigned long
best_copy_from_user_final(void *to, const void *from, unsigned long n)
{
	if (access_ok(VERIFY_READ, from, n)) {
		__copy_user_zeroing(to,from,n);
	}
	return n;
}

unsigned long
__best_copy_to_user(void *to, const void *from, unsigned long n)
{
	int BAR = (int)__builtin_return_address(0);
	int *caller = (int *)BAR - 1;
	if(boot_cpu_data.enable_fixups) {
	    if ( (boot_cpu_data.mmu_cr4_features & X86_CR4_OSFXSR) &&
		 (boot_cpu_data.x86_capability & X86_FEATURE_XMM) ) {
			*caller = (int)__kni_copy_to_user_nocheck - BAR;
			return(__kni_copy_to_user_nocheck(to, from, n));
		} else {
			*caller = (int)__best_copy_to_user_final - BAR;
			return(__best_copy_to_user_final(to, from, n));
		}
	} else {
		__copy_user(to,from,n);
		return n;
	}
}

static unsigned long
__best_copy_to_user_final(void *to, const void *from, unsigned long n)
{
	__copy_user(to,from,n);
	return n;
}

unsigned long
__best_copy_from_user(void *to, const void *from, unsigned long n)
{
	int BAR = (int)__builtin_return_address(0);
	int *caller = (int *)BAR - 1;
	if(boot_cpu_data.enable_fixups) {
	    if ( (boot_cpu_data.mmu_cr4_features & X86_CR4_OSFXSR) &&
		 (boot_cpu_data.x86_capability & X86_FEATURE_XMM) ) {
			*caller = (int)__kni_copy_from_user_nocheck - BAR;
			return(__kni_copy_from_user_nocheck(to, from, n));
		} else {
			*caller = (int)__best_copy_from_user_final - BAR;
			return(__best_copy_from_user_final(to, from, n));
		}
	} else {
		__copy_user_zeroing(to,from,n);
		return n;
	}
}

static unsigned long
__best_copy_from_user_final(void *to, const void *from, unsigned long n)
{
	__copy_user_zeroing(to,from,n);
	return n;
}

