/*
 * SIMD functions.  These replace the functions in asm-i386/string.h
 * whenever it makes sense.  These also un-inline those functions.
 *
 * Copyright 1999, Doug Ledford <dledford@redhat.com>
 *
 * These functions are simple and trivial, consider them to be
 * public domain
 */

#include <linux/string.h>
#include <linux/sched.h>
#include <asm/processor.h>
#include <asm/uaccess.h>
#include <asm/i387.h>

extern void * kni_memcpy(void * to, const void * from, size_t n)
{
	void *ret=to;
	size_t size;
	char xmm_space[32];
	int recursive;

	kernel_take_fpu_kni(recursive,2,&xmm_space[0],NULL);

	/*
	 * Align the destination on a 16byte boundary.
	 * The source doesn't have to be aligned.
	 */
	if ( (unsigned long)to & 0xf ) {
		size = 0x10 - ((unsigned long)to & 0xf);
		__memcpy(to, from, size);
		n -= size;
		from += size;
		to += size;
	}
	/*
	 * Prefetch the first two cachelines now.
	 */
	__asm__ __volatile__("prefetchnta 0x00(%0)\n\t"
			     "prefetchnta 0x20(%0)\n\t"
			     :
			     : "r" (from));
	/*
	 * Copy 32 bytes at a time.  The single unroll is good
	 * for a 30% performance boost in the copy.  Additional
	 * unrolls are not productive.  We are guaranteed to
	 * have at least 32 bytes of data to copy since the
	 * macro in string.h doesn't call into this function
	 * with less than 64 bytes of copy and we lost < 32
	 * bytes to alignment earlier.
	 */
	while (n >= 0x20) {
		__asm__ __volatile__(
				     "movups 0x00(%0),%%xmm0\n\t"
				     "movups 0x10(%0),%%xmm1\n\t"
				     "movntps %%xmm0,0x00(%1)\n\t"
				     "movntps %%xmm1,0x10(%1)\n\t"
				     : 
				     : "r" (from), "r" (to)
				     : "memory");
		from += 0x20;
		/*
		 * Note: Intermixing the prefetch at *exactly* this point
		 * in time has been shown to be the fastest possible.
		 * Timing these prefetch instructions is a complete black
		 * art with nothing but trial and error showing the way.
		 * To that extent, this optimum version was found by using
		 * a userland version of this routine that we clocked for
		 * lots of runs.  We then fiddled with ordering until we
		 * settled on our highest speen routines.  So, the long
		 * and short of this is, don't mess with instruction ordering
		 * here or suffer permance penalties you will.
		 */
		__asm__ __volatile__(
				     "prefetchnta 0x20(%0)\n\t"
				     : 
				     : "r" (from));
		to += 0x20;
		n -= 0x20;
	}
	if(n & 0x10) {
		__asm__ __volatile__("movups 0x00(%0),%%xmm0\n\t"
				     "movntps %%xmm0,0x00(%1)\n\t"
				     : 
				     : "r" (from), "r" (to)
				     : "memory");
		from += 0x10;
		to += 0x10;
		n -= 0x10;
	}
	if (n) {
		__memcpy(to, from, n);
	}
	SFENCE();
	kernel_release_fpu_kni(recursive,2,&xmm_space[0],NULL);
	return(ret);
}

extern void * kni_memset(void * s, char c, size_t count)
{
	size_t size;
	void *ret=s;
	char xmm_space[16];
	int recursive;

	kernel_take_fpu_kni(recursive,1,&xmm_space[0],NULL);
	/*
	 * align the destination on a 16 byte boundary
	 */
	size = (0x10 - ((unsigned long)s & 0xf));
	if(size != 0x10) {
		__memset_generic(s, c, size + 0x10);
	} else {
		__memset_generic(s, c, size);
	}
	/*
	 * We aligned the destination and also made sure there was at
	 * least 16 bytes of memory already set so that we could simply
	 * load that data into our XMM register to initialize it.
	 */
	__asm__ __volatile__("movups (%0), %%xmm0"
			     :
			     : "r" (s));
	count -= size;
	s += size;

	/*
	 * Do the copy by plopping out the register to memory.
	 * Note: Unrolling this was *totally* unproductive.  My benchmark
	 * showed that one or two plops per iteration produced the same
	 * speed to within .06 MByte/s of speed.  Considering that the
	 * routine benchmarked at over 3000 MByte/s, .06 is not statistically
	 * significant and only doing one drop per loop simplifies 
	 * overhead of book keeping.
	 */
	while(count & ~0xf) {
		__asm__ __volatile__("movntps %%xmm0,0x00(%0)\n\t"
				     :
				     : "r" (s));
		s += 0x10;
		count -= 0x10;
	}
	/*
	 * Catch any tailings...
	 */
	if(count) {
		__memset_generic(s, c, count);
	}
	SFENCE();
	kernel_release_fpu_kni(recursive,1,&xmm_space[0],NULL);
	return(ret);
}

#define __kni_copy_to_user(to,from,size)				\
do {									\
	int __d0, __d1;							\
	__asm__ __volatile__(						\
		"	prefetchnta 0x00(%2)\n"				\
		"	prefetchnta 0x20(%2)\n"				\
		"	jmp 200f\n"					\
		"100:	movups 0x00(%2),%%xmm0\n"			\
		"	movups 0x10(%2),%%xmm1\n"			\
		"1:	movntps %%xmm0,0x00(%1)\n"			\
		"2:	movntps %%xmm1,0x10(%1)\n"			\
		"	addl $0x20,%2\n"				\
		"	prefetchnta 0x20(%2)\n"				\
		"	addl $0x20,%1\n"				\
		"	subl $0x20,%0\n"				\
		"200:	cmpl $0x1f,%0\n"				\
		"	ja 100b\n"					\
		"	cmpl $0xf,%0\n"					\
		"	jbe 300f\n"					\
		"	movups 0x00(%2),%%xmm0\n"			\
		"3:	movntps %%xmm0,0x00(%1)\n"			\
		"	addl $0x10,%2\n"				\
		"	addl $0x10,%1\n"				\
		"	subl $0x10,%0\n"				\
		"300:\n"						\
		".section .fixup,\"ax\"\n"				\
		"6:	jmp 300b\n"					\
		"7:	addl $0x10,%1\n"				\
		"	addl $0x10,%2\n"				\
		"	subl $0x10,%0\n"				\
		"	jmp 300b\n"					\
		".previous\n"						\
		".section __ex_table,\"a\"\n"				\
		"	.align 4\n"					\
		"	.long 1b,6b\n"					\
		"	.long 2b,7b\n"					\
		"	.long 3b,6b\n"					\
		".previous"						\
		: "=&c"(size), "=r" (__d0), "=r" (__d1)		\
		: "0"(size), "1"(to), "2"(from)				\
		: "memory");						\
} while (0)

#define __kni_copy_from_user(to,from,size)				\
do {									\
	int __d0, __d1, tmp;						\
	__asm__ __volatile__(						\
		"	prefetchnta 0x00(%2)\n"				\
		"	prefetchnta 0x20(%2)\n"				\
		"	jmp 100f\n"					\
		"1:	movups 0x00(%2),%%xmm0\n"			\
		"2:	movups 0x10(%2),%%xmm1\n"			\
		"	movntps %%xmm0,0x00(%1)\n"			\
		"	movntps %%xmm1,0x10(%1)\n"			\
		"	addl $0x20,%2\n"				\
		"	prefetchnta 0x20(%2)\n"				\
		"	addl $0x20,%1\n"				\
		"	subl $0x20,%0\n"				\
		"100:	cmpl $0x1f,%0\n"				\
		"	ja 1b\n"					\
		"	cmpl $0xf,%0\n"					\
		"	jbe 200f\n"					\
		"3:	movups 0x00(%2),%%xmm0\n"			\
		"	movntps %%xmm0,0x00(%1)\n"			\
		"	addl $0x10,%2\n"				\
		"	addl $0x10,%1\n"				\
		"	subl $0x10,%0\n"				\
		"200:\n"						\
		".section .fixup,\"ax\"\n"				\
		"6:	addl $0x10,%1\n"				\
		"	subl $0x10,%0\n"				\
		"7:	xorps %%xmm0,%%xmm0\n"				\
		"	movl %0,%3\n"					\
		"	jmp 500f\n"					\
		"400:	movntps %%xmm0,0x00(%1)\n"			\
		"	addl $0x10,%1\n"				\
		"	subl $0x10,%3\n"				\
		"500:	cmpl $0x0f,%3\n"				\
		"	ja 400b\n"					\
		"	xorl %2,%2\n"					\
		"700:	testl %3,%3\n"					\
		"	je 200b\n"					\
		"	movb %2,(%1)\n"					\
		"	inc %1\n"					\
		"	dec %3\n"					\
		"	jmp 700b\n"					\
		".previous\n"						\
		".section __ex_table,\"a\"\n"				\
		"	.align 4\n"					\
		"	.long 1b,7b\n"					\
		"	.long 2b,6b\n"					\
		"	.long 3b,7b\n"					\
		".previous"						\
		: "=&c"(size), "=r" (__d0), "=q" (__d1), "=r"(tmp)	\
		: "0"(size), "1"(to), "2"(from)				\
		: "memory");						\
} while (0)


unsigned long
__kni_copy_to_user_nocheck(void *to, const void *from, unsigned long n)
{
	unsigned long size, count;
	char xmm_space[32];
	char xmm_reg_space[64];
	int recursive;

	if((unsigned long)to & 0xf) {
		count = size = 0x10 - ((unsigned long)to & 0xf);
		__copy_user(to,from,size);
		if(size)
			return(n + size);
		n -= count;
		to += count;
		from += count;
	}
	count = n;
	kernel_take_fpu_kni(recursive,2,&xmm_space[0],&xmm_reg_space[0]);
	__kni_copy_to_user(to,from,n);
	if(n && n < 0x10) {
		to += (count - n);
		from += (count - n);
		__copy_user(to,from,n);
	}
	SFENCE();
	kernel_release_fpu_kni(recursive,2,&xmm_space[0],&xmm_reg_space[0]);
	return n;
}

unsigned long
__kni_copy_from_user_nocheck(void *to, const void *from, unsigned long n)
{
	unsigned long size, count;
	char xmm_space[32];
	char xmm_reg_space[64];
	int recursive;

	if((unsigned long)to & 0xf) {
		count = size = 0x10 - ((unsigned long)to & 0xf);
		__copy_user_zeroing(to,from,size);
		if(size)
			return(n + size);
		n -= count;
		to += count;
		from += count;
	}
	count = n;
	kernel_take_fpu_kni(recursive,2,&xmm_space[0],&xmm_reg_space[0]);
	__kni_copy_from_user(to,from,n);
	if(n && n < 0x10) {
		to += (count - n);
		from += (count - n);
		__copy_user_zeroing(to,from,n);
	}
	SFENCE();
	kernel_release_fpu_kni(recursive,2,&xmm_space[0],&xmm_reg_space[0]);
	return n;
}



unsigned long
kni_copy_to_user(void *to, const void *from, unsigned long n)
{
	unsigned long size, count;
	char xmm_space[32];
	char xmm_reg_space[64];
	int recursive;

	if (access_ok(VERIFY_WRITE, to, n)) {
		if((unsigned long)to & 0xf) {
			count = size = 0x10 - ((unsigned long)to & 0xf);
			__copy_user(to,from,size);
			if(size)
				return(n + size);
			n -= count;
			to += count;
			from += count;
		}
		count = n;
		kernel_take_fpu_kni(recursive,2,&xmm_space[0],&xmm_reg_space[0]);
		__kni_copy_to_user(to,from,n);
		if(n && n < 0x10) {
			to += (count - n);
			from += (count - n);
			__copy_user(to,from,n);
		}
		SFENCE();
		kernel_release_fpu_kni(recursive,2,&xmm_space[0],&xmm_reg_space[0]);
	}
	return n;
}

unsigned long
kni_copy_from_user(void *to, const void *from, unsigned long n)
{
	unsigned long size, count;
	char xmm_space[32];
	char xmm_reg_space[64];
	int recursive;

	if (access_ok(VERIFY_READ, from, n)) {
		if((unsigned long)to & 0xf) {
			count = size = 0x10 - ((unsigned long)to & 0xf);
			__copy_user_zeroing(to,from,size);
			if(size)
				return(n + size);
			n -= count;
			to += count;
			from += count;
		}
		count = n;
		kernel_take_fpu_kni(recursive,2,&xmm_space[0],&xmm_reg_space[0]);
		__kni_copy_from_user(to,from,n);
		if(n && n < 0x10) {
			to += (count - n);
			from += (count - n);
			__copy_user_zeroing(to,from,n);
		}
		SFENCE();
		kernel_release_fpu_kni(recursive,2,&xmm_space[0],&xmm_reg_space[0]);
	}
	return n;
}


