/*
 * linux/drivers/block/ide-features.c
 *
 *  Copyright (C) 1999  Linus Torvalds & authors (see below)
 *  
 *  Andre Hedrick <andre@suse.com>
 *
 *  Extracts if ide.c to address the evolving transfer rate code for
 *  the SETFEATURES_XFER callouts.  Below are original authors of some or
 *  various parts of any given function below.
 *
 *  Mark Lord     <mlord@pobox.com>
 *  Gadi Oxman    <gadio@netvision.net.il>
 */

#define __NO_VERSION__
#include <linux/module.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/timer.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/major.h>
#include <linux/errno.h>
#include <linux/genhd.h>
#include <linux/malloc.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/hdreg.h>
#include <linux/ide.h>

#include <asm/byteorder.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/bitops.h>

/*
 * A Verbose noise maker for debugging on the attempted transfer rates.
 */
char *ide_xfer_verbose (byte xfer_rate)
{
	switch(xfer_rate) {
		case XFER_UDMA_7:	return("UDMA 7");
		case XFER_UDMA_6:	return("UDMA 6");
		case XFER_UDMA_5:	return("UDMA 5");
		case XFER_UDMA_4:	return("UDMA 4");
		case XFER_UDMA_3:	return("UDMA 3");
		case XFER_UDMA_2:	return("UDMA 2");
		case XFER_UDMA_1:	return("UDMA 1");
		case XFER_UDMA_0:	return("UDMA 0");
		case XFER_MW_DMA_2:	return("MW DMA 2");
		case XFER_MW_DMA_1:	return("MW DMA 1");
		case XFER_MW_DMA_0:	return("MW DMA 0");
		case XFER_SW_DMA_2:	return("SW DMA 2");
		case XFER_SW_DMA_1:	return("SW DMA 1");
		case XFER_SW_DMA_0:	return("SW DMA 0");
		case XFER_PIO_4:	return("PIO 4");
		case XFER_PIO_3:	return("PIO 3");
		case XFER_PIO_2:	return("PIO 2");
		case XFER_PIO_1:	return("PIO 1");
		case XFER_PIO_0:	return("PIO 0");
		case XFER_PIO_SLOW:	return("PIO SLOW");
		default:		return("XFER ERROR");
	}
}

/*
 *
 */
char *ide_media_verbose (ide_drive_t *drive)
{
	switch (drive->media) {
		case ide_scsi:		return("scsi   ");
		case ide_disk:		return("disk   ");
		case ide_optical:	return("optical");
		case ide_cdrom:		return("cdrom  ");
		case ide_tape:		return("tape   ");
		case ide_floppy:	return("floppy ");
		default:		return("???????");
	}
}

/*
 * A Verbose noise maker for debugging on the attempted dmaing calls.
 */
char *ide_dmafunc_verbose (ide_dma_action_t dmafunc)
{
	switch (dmafunc) {
		case ide_dma_read:		return("ide_dma_read");
		case ide_dma_write:		return("ide_dma_write");
		case ide_dma_begin:		return("ide_dma_begin");
		case ide_dma_end:		return("ide_dma_end:");
		case ide_dma_check:		return("ide_dma_check");
		case ide_dma_on:		return("ide_dma_on");
		case ide_dma_off:		return("ide_dma_off");
		case ide_dma_off_quietly:	return("ide_dma_off_quietly");
		case ide_dma_test_irq:		return("ide_dma_test_irq");
		case ide_dma_bad_drive:		return("ide_dma_bad_drive");
		case ide_dma_good_drive:	return("ide_dma_good_drive");
		case ide_dma_verbose:		return("ide_dma_verbose");
		case ide_dma_retune:		return("ide_dma_retune");
		case ide_dma_lostirq:		return("ide_dma_lostirq");
		case ide_dma_timeout:		return("ide_dma_timeout");
		default:			return("unknown");
	}
}

char * ide_ioctl_verbose (unsigned int cmd)
{
	switch (cmd) {
		case HDIO_SET_PIO_MODE:		return("HDIO_SET_PIO_MODE");
		case HDIO_SET_32BIT:		return("HDIO_SET_32BIT");
		case HDIO_GET_32BIT:		return("HDIO_GET_32BIT");
		case HDIO_SET_MULTCOUNT:	return("HDIO_SET_MULTCOUNT");
		case HDIO_GET_MULTCOUNT:	return("HDIO_GET_MULTCOUNT");
		case HDIO_SET_UNMASKINTR:	return("HDIO_SET_UNMASKINTR");
		case HDIO_GET_UNMASKINTR:	return("HDIO_GET_UNMASKINTR");
		case HDIO_SET_DMA:		return("HDIO_SET_DMA");
		case HDIO_GET_DMA:		return("HDIO_GET_DMA");
		case HDIO_SET_NOWERR:		return("HDIO_SET_NOWERR");
		case HDIO_GET_NOWERR:		return("HDIO_GET_NOWERR");
		case HDIO_SET_KEEPSETTINGS:	return("HDIO_SET_KEEPSETTINGS");
		case HDIO_GET_KEEPSETTINGS:	return("HDIO_GET_KEEPSETTINGS");

		case HDIO_GETGEO:		return("HDIO_GETGEO");
		case HDIO_GETGEO_BIG:		return("HDIO_GETGEO_BIG");
		case HDIO_GETGEO_BIG_RAW:	return("HDIO_GETGEO_BIG_RAW");
		case BLKRASET:			return("BLKRASET");
		case BLKGETSIZE:		return("BLKGETSIZE");
		case BLKRRPART:			return("BLKRRPART");
		case HDIO_OBSOLETE_IDENTITY:	return("HDIO_OBSOLETE_IDENTITY");
		case HDIO_GET_IDENTITY:		return("HDIO_GET_IDENTITY");
		case HDIO_DRIVE_CMD:		return("HDIO_DRIVE_CMD");
		case HDIO_DRIVE_TASK:		return("HDIO_DRIVE_TASK");
		case HDIO_SCAN_HWIF:		return("HDIO_SCAN_HWIF");
		case HDIO_UNREGISTER_HWIF:	return("HDIO_UNREGISTER_HWIF");
		case HDIO_SET_NICE:		return("HDIO_SET_NICE");
		case HDIO_DRIVE_RESET:		return("HDIO_DRIVE_RESET");
/*		case HDIO_TRISTATE_HWIF:	return("HDIO_TRISTATE_HWIF");	*/
		case BLKROSET:			return("BLKROSET");
		case BLKROGET:			return("BLKROGET");
		case BLKFLSBUF:			return("BLKFLSBUF");
		case BLKSSZGET:			return("BLKSSZGET");
		case BLKELVGET:			return("BLKELVGET");
		case BLKELVSET:			return("BLKELVSET");
		default:			return("unknown");
	}
}

char * ide_task_cmd_verbose (byte task)
{
	switch(task) {
		case CFA_TRANSLATE_SECTOR:		return("CFA_TRANSLATE_SECTOR");
		case WIN_IDENTIFY:			return("WIN_IDENTIFY");
		case WIN_PIDENTIFY:			return("WIN_PIDENTIFY");
		case WIN_READ:				return("WIN_READ");
		case WIN_READ_BUFFER:			return("WIN_READ_BUFFER");
		case WIN_WRITE:				return("WIN_WRITE");
		case WIN_WRITE_VERIFY:			return("WIN_WRITE_VERIFY");
		case WIN_WRITE_BUFFER:			return("WIN_WRITE_BUFFER");
		case CFA_WRITE_SECT_WO_ERASE:		return("CFA_WRITE_SECT_WO_ERASE");
		case WIN_DOWNLOAD_MICROCODE:		return("WIN_DOWNLOAD_MICROCODE");
		case WIN_MULTREAD:			return("WIN_MULTREAD");
		case CFA_WRITE_MULTI_WO_ERASE:		return("CFA_WRITE_MULTI_WO_ERASE");
		case WIN_MULTWRITE:			return("WIN_MULTWRITE");
		case WIN_SECURITY_DISABLE:		return("WIN_SECURITY_DISABLE");
		case WIN_SECURITY_ERASE_UNIT:		return("WIN_SECURITY_ERASE_UNIT");
		case WIN_SECURITY_SET_PASS:		return("WIN_SECURITY_SET_PASS");
		case WIN_SECURITY_UNLOCK:		return("WIN_SECURITY_UNLOCK");
		case WIN_SMART:				return("WIN_SMART");
#ifdef CONFIG_BLK_DEV_IDEDMA
		case WIN_READDMA:			return("WIN_READDMA");
		case WIN_IDENTIFY_DMA:			return("WIN_IDENTIFY_DMA");
		case WIN_READDMA_QUEUED:		return("WIN_READDMA_QUEUED");
		case WIN_WRITEDMA:			return("WIN_WRITEDMA");
		case WIN_WRITEDMA_QUEUED:		return("WIN_WRITEDMA_QUEUED");
#endif
		case WIN_SETFEATURES:			return("WIN_SETFEATURES");
		case WIN_NOP:				return("WIN_NOP");
		case CFA_REQ_EXT_ERROR_CODE:		return("CFA_REQ_EXT_ERROR_CODE");
		case CFA_ERASE_SECTORS:			return("CFA_ERASE_SECTORS");
		case WIN_VERIFY:			return("WIN_VERIFY");
		case WIN_SEEK:				return("WIN_SEEK");
		case WIN_SPECIFY:			return("WIN_SPECIFY");
		case WIN_FLUSH_CACHE:			return("WIN_FLUSH_CACHE");
		case WIN_STANDBYNOW1:			return("WIN_STANDBYNOW1");
		case WIN_STANDBYNOW2:			return("WIN_STANDBYNOW2");
		case WIN_SLEEPNOW1:			return("WIN_SLEEPNOW1");
		case WIN_SLEEPNOW2:			return("WIN_SLEEPNOW2");
		case WIN_SETIDLE1:			return("WIN_SETIDLE1");
		case DISABLE_SEAGATE:			return("DISABLE_SEAGATE");
		case WIN_CHECKPOWERMODE1:		return("WIN_CHECKPOWERMODE1");
		case WIN_CHECKPOWERMODE2:		return("WIN_CHECKPOWERMODE2");
		case WIN_GETMEDIASTATUS:		return("WIN_GETMEDIASTATUS");
		case WIN_MEDIAEJECT:			return("WIN_MEDIAEJECT");
		case WIN_SETMULT:			return("WIN_SETMULT");
		case WIN_READ_NATIVE_MAX:		return("WIN_READ_NATIVE_MAX");
		case WIN_SET_MAX:			return("WIN_SET_MAX");
		case WIN_SECURITY_ERASE_PREPARE:	return("WIN_SECURITY_ERASE_PREPARE");
		case WIN_SECURITY_FREEZE_LOCK:		return("WIN_SECURITY_FREEZE_LOCK");
		case EXABYTE_ENABLE_NEST:		return("EXABYTE_ENABLE_NEST");
		case WIN_DOORLOCK:			return("WIN_DOORLOCK");
		case WIN_DOORUNLOCK:			return("WIN_DOORUNLOCK");
		case WIN_RESTORE:			return("WIN_RESTORE");
		case WIN_FORMAT:			return("WIN_FORMAT");
		case WIN_INIT:				return("WIN_INIT");
		case WIN_DIAGNOSE:			return("WIN_DIAGNOSE");
		case WIN_DEVICE_RESET:			return("WIN_DEVICE_RESET");
		case WIN_QUEUED_SERVICE:		return("WIN_QUEUED_SERVICE");
		case WIN_PACKETCMD:			return("WIN_PACKETCMD");
		default:				return("unknown");
	}
}

/*
 *
 */
byte ide_auto_reduce_xfer (ide_drive_t *drive)
{
	if (!drive->crc_count)
		return drive->current_speed;
	drive->crc_count = 0;

	switch(drive->current_speed) {
		case XFER_UDMA_7:	return XFER_UDMA_6;
		case XFER_UDMA_6:	return XFER_UDMA_5;
		case XFER_UDMA_5:	return XFER_UDMA_4;
		case XFER_UDMA_4:	return XFER_UDMA_3;
		case XFER_UDMA_3:	return XFER_UDMA_2;
		case XFER_UDMA_2:	return XFER_UDMA_1;
		case XFER_UDMA_1:	return XFER_UDMA_0;
		case XFER_UDMA_0:
			if (drive->id->dma_mword & 0x0004) return XFER_MW_DMA_2;
			else if (drive->id->dma_mword & 0x0002) return XFER_MW_DMA_1;
			else if (drive->id->dma_mword & 0x0001) return XFER_MW_DMA_0;
			else return XFER_PIO_4;
		case XFER_MW_DMA_2:	return XFER_MW_DMA_1;
		case XFER_MW_DMA_1:	return XFER_MW_DMA_0;
		case XFER_MW_DMA_0:
			if (drive->id->dma_1word & 0x0004) return XFER_SW_DMA_2;
			else if (drive->id->dma_1word & 0x0002) return XFER_SW_DMA_1;
			else if (drive->id->dma_1word & 0x0001) return XFER_SW_DMA_0;
			else return XFER_PIO_4;
		case XFER_SW_DMA_2:	return XFER_SW_DMA_1;
		case XFER_SW_DMA_1:	return XFER_SW_DMA_0;
		case XFER_SW_DMA_0:
			{
				return XFER_PIO_4;
			}
		case XFER_PIO_4:	return XFER_PIO_3;
		case XFER_PIO_3:	return XFER_PIO_2;
		case XFER_PIO_2:	return XFER_PIO_1;
		case XFER_PIO_1:	return XFER_PIO_0;
		case XFER_PIO_0:
		default:		return XFER_PIO_SLOW;
	}
}

int ide_driveid_update (ide_drive_t *drive)
{
	/*
	 * Re-read drive->id for possible DMA mode
	 * change (copied from ide-probe.c)
	 */
	struct hd_driveid *id;
	unsigned long irqs, flags;
	long timeout;

	probe_irq_off(probe_irq_on());
	irqs = probe_irq_on();
	SELECT_MASK(HWIF(drive), drive, 1);
	if (IDE_CONTROL_REG)
		OUT_BYTE(drive->ctl,IDE_CONTROL_REG);
	ide_delay_50ms();
	OUT_BYTE(WIN_IDENTIFY, IDE_COMMAND_REG);
	timeout = IDE_TIMEOUT_50MS(WAIT_WORSTCASE);
	do {
		if (timeout-- < 1) {
			if (irqs)
				(void) probe_irq_off(irqs);
			SELECT_MASK(HWIF(drive), drive, 0);
			return 0;	/* drive timed-out */
		}
		ide_delay_50ms();	/* give drive a breather */
	} while (IN_BYTE(IDE_ALTSTATUS_REG) & BUSY_STAT);
	ide_delay_50ms();	/* wait for IRQ and DRQ_STAT */
	if (!OK_STAT(GET_STAT(),DRQ_STAT,BAD_R_STAT)) {
		SELECT_MASK(HWIF(drive), drive, 0);
		return 0;
	}
	__save_flags(flags);	/* local CPU only */
	__cli();		/* local CPU only; some systems need this */
	SELECT_MASK(HWIF(drive), drive, 0);
	id = kmalloc(SECTOR_WORDS*4, GFP_ATOMIC);
	ide_input_data(drive, id, SECTOR_WORDS);
	(void) GET_STAT();	/* clear drive IRQ */
	ide__sti();		/* local CPU only */
	__restore_flags(flags);	/* local CPU only */
	ide_fix_driveid(id);
	if (id) {
		drive->id->dma_ultra = id->dma_ultra;
		drive->id->dma_mword = id->dma_mword;
		drive->id->dma_1word = id->dma_1word;
		/* anything more ? */
		kfree(id);
	}
	return 1;
}


/*
 * Verify that we are doing an approved SETFEATURES_XFER with respect
 * to the hardware being able to support request.  Since some hardware
 * can improperly report capabilties, we check to see if the host adapter
 * in combination with the device (usually a disk) properly detect
 * and acknowledge each end of the ribbon.
 */
int ide_ata66_check (ide_drive_t *drive, byte cmd, byte nsect, byte feature)
{
	if ((cmd == WIN_SETFEATURES) &&
	    (nsect > XFER_UDMA_2) &&
	    (feature == SETFEATURES_XFER)) {
		if (!HWIF(drive)->udma_four) {
			printk("%s: Speed warnings UDMA 3/4/5 is not functional.\n", HWIF(drive)->name);
			return 1;
		}
#ifndef CONFIG_IDEDMA_IVB
		if ((drive->id->hw_config & 0x6000) == 0) {
#else /* !CONFIG_IDEDMA_IVB */
		if ((drive->id->hw_config & 0x2000) == 0) {
#endif /* CONFIG_IDEDMA_IVB */
			printk("%s: Speed warnings UDMA 3/4/5 is not functional.\n", drive->name);
			return 1;
		}
	}
	return 0;
}

/*
 * Backside of HDIO_DRIVE_CMD call of SETFEATURES_XFER.
 * 1 : Safe to update drive->id DMA registers.
 * 0 : OOPs not allowed.
 */
int set_transfer (ide_drive_t *drive, byte cmd, byte nsect, byte feature)
{
	if ((cmd == WIN_SETFEATURES) &&
	    (nsect >= XFER_SW_DMA_0) &&
	    (feature == SETFEATURES_XFER) &&
	    (drive->id->dma_ultra ||
	     drive->id->dma_mword ||
             drive->id->dma_1word))
		return 1;
	return 0;
}

/*
 *  All hosts that use the 80c ribbon mus use!
 */
byte eighty_ninty_three (ide_drive_t *drive)
{
	return ((byte) ((HWIF(drive)->udma_four) &&
#ifndef CONFIG_IDEDMA_IVB
			(drive->id->hw_config & 0x4000) &&
#endif /* CONFIG_IDEDMA_IVB */
			(drive->id->hw_config & 0x2000)) ? 1 : 0);
}

/*
 * Similar to ide_wait_stat(), except it never calls ide_error internally.
 * This is a kludge to handle the new ide_config_drive_speed() function,
 * and should not otherwise be used anywhere.  Eventually, the tuneproc's
 * should be updated to return ide_startstop_t, in which case we can get
 * rid of this abomination again.  :)   -ml
 *
 * It is gone..........
 *
 * const char *msg == consider adding for verbose errors.
 */
int ide_config_drive_speed (ide_drive_t *drive, byte speed)
{
	ide_hwif_t *hwif = HWIF(drive);
	int	i, error = 1;
	byte stat;

#ifdef CONFIG_BLK_DEV_IDEDMA
	byte unit = (drive->select.b.unit & 0x01);
	outb(inb(hwif->dma_base+2) & ~(1<<(5+unit)), hwif->dma_base+2);
#endif /* CONFIG_BLK_DEV_IDEDMA */

	/*
	 * Don't use ide_wait_cmd here - it will
	 * attempt to set_geometry and recalibrate,
	 * but for some reason these don't work at
	 * this point (lost interrupt).
	 */
        /*
         * Select the drive, and issue the SETFEATURES command
         */
	disable_irq(hwif->irq);	/* disable_irq_nosync ?? */
	udelay(1);
	SELECT_DRIVE(HWIF(drive), drive);
	SELECT_MASK(HWIF(drive), drive, 0);
	udelay(1);
	if (IDE_CONTROL_REG)
		OUT_BYTE(drive->ctl | 2, IDE_CONTROL_REG);
	OUT_BYTE(speed, IDE_NSECTOR_REG);
	OUT_BYTE(SETFEATURES_XFER, IDE_FEATURE_REG);
	OUT_BYTE(WIN_SETFEATURES, IDE_COMMAND_REG);
	if ((IDE_CONTROL_REG) && (drive->quirk_list == 2))
		OUT_BYTE(drive->ctl, IDE_CONTROL_REG);
	udelay(1);
	/*
	 * Wait for drive to become non-BUSY
	 */
	if ((stat = GET_STAT()) & BUSY_STAT) {
		unsigned long flags;
		long timeout;

		__save_flags(flags);	/* local CPU only */
		ide__sti();		/* local CPU only -- for jiffies */
		timeout = IDE_TIMEOUT_MS(WAIT_CMD);
		while ((stat = GET_STAT()) & BUSY_STAT) {
			if (timeout-- < 1)
				break;
			mdelay(1);
		}
		__restore_flags(flags); /* local CPU only */
	}

	/*
	 * Allow status to settle, then read it again.
	 * A few rare drives vastly violate the 400ns spec here,
	 * so we'll wait up to 10usec for a "good" status
	 * rather than expensively fail things immediately.
	 * This fix courtesy of Matthew Faupel & Niccolo Rigacci.
	 */
	for (i = 0; i < 10; i++) {
		udelay(1);
		if (OK_STAT((stat = GET_STAT()), DRIVE_READY, BUSY_STAT|DRQ_STAT|ERR_STAT)) {
			error = 0;
			break;
		}
	}

	SELECT_MASK(HWIF(drive), drive, 0);

	enable_irq(hwif->irq);

	if (error) {
		(void) ide_dump_status(drive, "set_drive_speed_status", stat);
		return error;
	}

	drive->id->dma_ultra &= ~0xFF00;
	drive->id->dma_mword &= ~0x0F00;
	drive->id->dma_1word &= ~0x0F00;

#ifdef CONFIG_BLK_DEV_IDEDMA
	if (speed > XFER_PIO_4) {
		outb(inb(hwif->dma_base+2)|(1<<(5+unit)), hwif->dma_base+2);
	} else {
		outb(inb(hwif->dma_base+2) & ~(1<<(5+unit)), hwif->dma_base+2);
	}
#endif /* CONFIG_BLK_DEV_IDEDMA */

	switch(speed) {
		case XFER_UDMA_7:   drive->id->dma_ultra |= 0x8080; break;
		case XFER_UDMA_6:   drive->id->dma_ultra |= 0x4040; break;
		case XFER_UDMA_5:   drive->id->dma_ultra |= 0x2020; break;
		case XFER_UDMA_4:   drive->id->dma_ultra |= 0x1010; break;
		case XFER_UDMA_3:   drive->id->dma_ultra |= 0x0808; break;
		case XFER_UDMA_2:   drive->id->dma_ultra |= 0x0404; break;
		case XFER_UDMA_1:   drive->id->dma_ultra |= 0x0202; break;
		case XFER_UDMA_0:   drive->id->dma_ultra |= 0x0101; break;
		case XFER_MW_DMA_2: drive->id->dma_mword |= 0x0404; break;
		case XFER_MW_DMA_1: drive->id->dma_mword |= 0x0202; break;
		case XFER_MW_DMA_0: drive->id->dma_mword |= 0x0101; break;
		case XFER_SW_DMA_2: drive->id->dma_1word |= 0x0404; break;
		case XFER_SW_DMA_1: drive->id->dma_1word |= 0x0202; break;
		case XFER_SW_DMA_0: drive->id->dma_1word |= 0x0101; break;
		default: break;
	}
	return error;
}

EXPORT_SYMBOL(ide_auto_reduce_xfer);
EXPORT_SYMBOL(ide_driveid_update);
EXPORT_SYMBOL(ide_ata66_check);
EXPORT_SYMBOL(set_transfer);
EXPORT_SYMBOL(eighty_ninty_three);
EXPORT_SYMBOL(ide_config_drive_speed);
