/* DP83815.c -- National Semiconductor DP83815 PCI Ethernet Controller Driver */
/* $Id: dp83815.c,v 1.34 2001/08/14 01:23:49 asun Exp $ */

static const char *version = "dp83815.c:v1.30 National Semiconductor DP83815 PCI Ethernet Driver\n";

/*	
1. INTRODUCTION
---------------

The DP83815 is a PCI bus based single chip 10/100 Mbps ethernet
controller targeted for PC mother boards, adapter cards, and embedded
systems. It fully implements the V2.2 33MHz PCI bus interface for host
communications with power management support. Packet descriptors and
data are transferred via bus-mastering, reducing the burden on the
host CPU. The DP83815 can support full duplex 10/100 Mbps transmission
and reception. This driver has been updated to support rev C of the chip.

The DP83815 driver is developed for use on Linux v2.0.X and v2.2.X and
has been tested with 2.0.36 and 2.2.13 versions of the kernel.It can be
configured either as a static driver into the kernel or as a loadable
module. The driver is capable of 10/100 Mbps Full/Half duplex mode
operation via an internal PHY. Detailed features are given below.

2.1 Chip Features

  o The chip is bus-master capable and transfers packet descriptors
    and data using DMA with burst sizes upto 128 words.

  o It supports 10 Mbps Ethernet and 100 Mbps Fast Ethernet via
    an internal PHY and emerging 1-2 Mbps home networking
    solutions via external PHY.

  o Flexible Rx packet filtration including: perfect match, broadcast, 
    multicast/unicast hash table, deep packet pattern matching.

  o Internal Tx and Rx data FIFOs of 2KB each

  o 802.3u Auto-Negotiation - advertised features configurable via
    EEPROM

  o Full duplex support for 10 and 100Mb/s data rates

2.2 Driver Features

2.2.1 Supported Features

  o 10 Mbps full duplex and half duplex
  o 100 Mbps full duplex and half duplex
  o Auto Negotiation
  o Broadcast and perfect match transmit and receive
  o ARP transmit and receive
  o Promiscuous mode
  o Multicast support
  o Loadable Kernel Module
  o PCI bus probe for auto-configuration when configured as a static
     driver or a loadable kernel module
  o Endian, and CPU architecture neutral
  o Supports rev C of the chip
  o 2.0.xx and 2.2.xx Linux kernels
 
2.2.2 Unsupported Features

  o ACPI
  o Multiple ethernet addresses 
  o Wake on LAN and OnNow support for PC98/99, and other power-management
    features.
  o Magic packets with SecureON, VLAN
  o Programming EEPROM/Flash
  o Old Linux kernels 1.xx.xx
  o Newer MP kernels 2.2.xx 
  o Remote boot           


3. DRIVER OPERATION
-------------------

Buffer management used in this version of the driver sets up the
transmit and receive buffer descriptors in a ring. The rings, shared
between the driver and the device, are located in memory buffers 
allocated in the kernel. The device accesses these buffers via DMA.

Each device will be allocated two similar queues of descriptors, one
for transmit and the other for receive. These queues are linked
together into a ring. For the default size of the transmit rings is
10, while the size of the receive ring is 30. These values can be
changed via macros in the driver source file.

The driver implements no copy transmits. All transmit descriptors are
created with no associated buffers, but the descriptor uses the
sk_buff passed to it via the transmit routine. The sk_buff is held by
the driver until the next transmit operation that follows the actual
transmit of the packet.

Each receive descriptor is initialized, at creation time, with an
sk_buff. Each sk_buff has a data buffer that can hold a max ethernet
packet. When a packet is suceesfully received, the skb is directly
sent to the stack with no copy, and it is replaced with a newly
allocated sk_buff. But if the packet less than 128 bytes
was received, the data is instead copied into a new skb. 

The driver makes of appropriate macros to adjust for endian-ness of
the host CPU, PCI bus, Ethernet, and the DP device. It also performs
appropriate translations of addresses for device use. These builtin
features makes the driver extremely portable to run on Linux systems
with non-Intel CPUs with minimum changes. 

The driver has a table, dp83815_hw_id_tbl[] to store the hardware_id
information for the DP83815 device from all Vendors. In this version
of the driver, the dp83815_hw_id_tbl has only one entry for the
vendor_id and board_id of the NS_DP83815. If the device's ID differs
from these values, the table should be extended with the new values.


4. FILES IN THIS RELEASE
------------------------

The following files are included in a tar ball, dp83815.tar.gz

  README.Linux   - This file
  dp83815.c      - Driver source file
  dp83815.h      - Driver header file
  dp83815.o      - Driver object module for 2.0.36 kernel
  Makefile       - Rules to compile and install the driver.
 

5. INSTALLATION
---------------

This version of the driver requires a base RedHat Linux ver 5.2 with 
linux kernel version 2.0.36 (or 2.2.13), basic utilities and the C development system.

To install the driver, login as root, and extract the files into an 
appropriate directory.

	# mkdir dp
	# cd dp
	# tar xvzf <pathToTheTarBall>/dp83815.tar.gz

To create a new driver object module, run

	# make

To install the current driver object module in the filesystem run,

	# make install

You'll need to edit the following files to configure the driver as a
valid ethernet device in the system

  o /etc/hosts
  o /etc/sysconfig/network-scripts/ifcfg-eth?
  o /etc/conf.modules

The driver can be dynamically loaded and unloaded from a running kernel 
with the following commands,

	# insmod dp83815
	# rmmod dp83815

To install and test the driver, shutdown the network stack, install
the driver, and re-start the startup the stack again.

	# sh /etc/rc.d/init.d/network stop
	# insmod dp83815
	# sh /etc/rc.d/init.d/network start


6. FUNCTIONALITY TESTING
------------------------

When the driver is loaded into the system via `insmod' it probes the
PCI bus to locate all DP83815 devices, and creates control structures
for each. The driver logs a couple of messages available in
`/var/log/messages' for each device with information about its PCI
geographic location, IRQ, IO address, and some basic debug information 
(addresses of some important structures).

All the devices on the PCI bus can be listed by,

	# cat /proc/pci

IRQ and IO address information from this can be correlated with the
information displayed by the driver in `/var/log/messages'

When the TCP/IP stack is initialized, it opens all configured ethernet 
devices, and initializes them for use. At this time, the driver will
perform autonegotiation and log information about the link status.
The driver can then be tested by running ping, telnet, ftp, NFS etc.


7. KNOWN DRIVER PROBLEMS
------------------------

o  Since the device requires that the receive buffer be aligned on a
   4-byte (word) boundary, it is not possible to align the IP header
   on a word boundary unless the received packet is copied. Since this
   version of the driver support a no-copy receive operation, the IP
   header will not be aligned. This is not a problem for Intel CPUs,
   but will cause exceptions on RISC CPUs (PowerPC, Alpha, ...). While
   porting the driver to these platforms, the no-copy receive can be
   turned off, but setting the DP_RX_COPY_THRESHOLD to ETH_MAX_PKT_SIZE,
   whereby forcing all receive buffers to be copied.

   The driver ensures that the IP headers in a copied buffer is aligned
   on a word-boundary.

8. REVISION HISTORY 
--------------------

v1.30 : support for Linux Kernel 2.2 was added

v1.20 : proper initialisation of TXCFG and RXCFG registers for full duplex
        mode was done.  
	in start_xmit, BUFPTR was updated prior to CMDSTS of the descriptor 
	infinite loop in case of dev_reset failure was corrected.

v1.10 : this is the first release of the driver

9. DOCUMENTATION REFERENCES
---------------------------

1. National Semiconductor DP83815 10/100 Mb/s Integrated PCI Ethernet
   Media Access Controller and Physical Layer, 1998.
2. Linux Device Drivers, Alessandro Rubini, 1998.


10. BUG REPORTS
--------------

Bug reports and enhancement requests should be directed to
<ethernet@teamF1.com>. Please include the driver version number,
hardware information, and Linux kernel version number along with the
bug report.

*/

/* Generic Kernel Module/Driver Headers */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h> 

#include <linux/pci.h>
#include <linux/ioport.h>
#include <asm/io.h>         /* IO stuff */
#include <asm/byteorder.h>
#include <linux/delay.h>

#ifdef CONFIG_COBALT_GEN_V
#include <linux/lcd.h>

static void dp83815_check_link(const void);
static struct lcd_handler dp83815_lcd_handler = { dp83815_check_link, 
                                                  &dp83815_lcd_handler,
                                                  &dp83815_lcd_handler };
#ifdef MODULE
#define DP_WOL_SUPPORT
#endif
#define LED_SET_TXRX(a) (lcd_fp_led_state(1, ((a)->name[3] == '0') ? \
                         LED_ETH0_TXRX : LED_ETH1_TXRX))
#define LED_SET_LINK(a, b) (lcd_fp_led_state((b), ((a)->name[3] == '0') ? \
                         LED_ETH0_LINK : LED_ETH1_LINK))
#else
#define LED_SET_TXRX(a)
#define LED_SET_LINK(a, b)
#endif

#ifdef DP_WOL_SUPPORT
#include <linux/notifier.h>
#include <linux/reboot.h>
#endif

/* Ethernet Driver Specific Headers */

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/malloc.h>
#include <linux/if_ether.h>

#include "dp83815.h"

/* passed in module options */
#define MAX_UNITS 8             /* More are supported, limit only on options */
static int options[MAX_UNITS] = {-1, -1, -1, -1, -1, -1, -1, -1};
static int full_duplex[MAX_UNITS] = {-1, -1, -1, -1, -1, -1, -1, -1};

MODULE_DESCRIPTION("National Semiconductor DP83815 series PCI Ethernet Driver");
MODULE_PARM(options, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(full_duplex, "1-" __MODULE_STRING(MAX_UNITS) "i");


/* Macros to make the driver compatible with both 2.0.x and 2.2.x kernels */ 

/* to convert Linux kernel version into hexadecimal number */
#define DP_KERNEL_VERSION(a,b,c)	(((a) << 16) + ((b) << 8) + (c))

/* For Linux Kernel 2.0.x */
#if ((LINUX_VERSION_CODE >= DP_KERNEL_VERSION (2,0,0)) &&                                                  (LINUX_VERSION_CODE < DP_KERNEL_VERSION (2,1,0))) 
#include <linux/bios32.h>	/* PCI BIOS support*/
int     dp_dev_cur = 0;        	/* current pci device count*/
#define DP_PCIBIOS_FIND_DEVICE	/* pcibios_find_device API is available */
#undef  DP_PCI_FIND_DEVICE	/* pci_find_device API is not available */
#define DP_KFREE_SKB(skb)		dev_kfree_skb (skb, FREE_WRITE)
#define DP_INW(io_port)			__inw (io_port)
#define DP_OUTW(data, io_port)		__outw (data, io_port)
#define DP_INL(io_port)			__inl (io_port)
#define DP_OUTL(data, io_port)		__outl (data, io_port)
#define DP_STATS_INCR(stats, mode, l)	((stats)->mode##_packets++)
  
/* for 2.2.X kernel */
#elif (LINUX_VERSION_CODE >= DP_KERNEL_VERSION (2,2,0)) 
struct  pci_dev *dp_pci_dev = NULL;	/* pci device structure */
#undef DP_PCIBIOS_FIND_DEVICE	/* pcibios_find_device API is not available */
#define DP_PCI_FIND_DEVICE	/* pci_find_device API is available */
#define DP_KFREE_SKB(skb)               dev_kfree_skb (skb)
#define DP_INW(io_port)                 inw (io_port)
#define DP_OUTW(data, io_port)		outw (data, io_port)
#define DP_INL(io_port)                 inl (io_port)
#define DP_OUTL(data, io_port)          outl (data, io_port)
#define DP_STATS_INCR(stats, mode, l)	do { \
						(stats)->mode##_packets++; \
						(stats)->mode##_bytes += (l); \
					} while (0)

#else /* for all other versions of the kernel */
#error "KERNEL VERSIONS OTHER THAN 2.0.X and 2.2.X ARE NOT SUPPORTED!!\n"
#endif 


/*
 * SWAP_BUS_TO_CPU_XX and SWAP_CPU_TO_BUS_XX macros swap 16 and 32 bit values
 * between the PCI bus' little endian byte-order and the CPU's native
 * byte-order.
 */
#ifdef __BIG_ENDIAN
#define BUS_TO_CPU_SWAP_16(X)    swap_16(X)
#define BUS_TO_CPU_SWAP_32(X)    swap_32(X)
#define CPU_TO_BUS_SWAP_16(X)    swap_16(X)
#define CPU_TO_BUS_SWAP_32(X)    swap_32(X)
#define CPU_TO_NET_SWAP_16(X)    ((u16)(X))
#define CPU_TO_NET_SWAP_32(X)    ((u32)(X))
#define NET_TO_CPU_SWAP_16(X)    ((u16)(X))
#define NET_TO_CPU_SWAP_32(X)    ((u32)(X))
#else
#define BUS_TO_CPU_SWAP_32(X)    ((u32)(X))
#define BUS_TO_CPU_SWAP_16(X)    ((u16)(X))
#define CPU_TO_BUS_SWAP_32(X)    ((u32)(X))
#define CPU_TO_BUS_SWAP_16(X)    ((u16)(X))
#define CPU_TO_NET_SWAP_16(X)    swap_16(X)
#define CPU_TO_NET_SWAP_32(X)    swap_32(X)
#define NET_TO_CPU_SWAP_16(X)    swap_16(X)
#define NET_TO_CPU_SWAP_32(X)    swap_32(X)
#endif

/* Macros translate addresses between PCI bus and CPU */
#define BUS_TO_CPU_ADDR_XLATE(X)  BUS_TO_CPU_SWAP_32(bus_to_virt(X))
#define CPU_TO_BUS_ADDR_XLATE(X)  CPU_TO_BUS_SWAP_32(virt_to_bus(X))

/* Macros to read/write 32/16 bit data to/from DP83815 device registers. */
#define DP_REG32_WRITE(reg, val)  io_write_32 (iobase+(reg),            \
                                               CPU_TO_BUS_SWAP_32(val))
#define DP_REG32_READ(reg)        BUS_TO_CPU_SWAP_32(io_read_32 (iobase+(reg)))
#define DP_REG16_WRITE(reg, val)  io_write_16 (iobase+(reg),            \
                                               CPU_TO_BUS_SWAP_16(val))
#define DP_REG16_READ(reg)        BUS_TO_CPU_SWAP_16(io_read_16 (iobase+(reg)))
#define DP_REG32_SET(reg, val)    DP_REG32_WRITE(reg,DP_REG32_READ(reg)|(val))
#define DP_REG32_CLR(reg, val)    DP_REG32_WRITE(reg,DP_REG32_READ(reg)&~(val))
#define DP_REG16_SET(reg, val)    DP_REG16_WRITE(reg,DP_REG16_READ(reg)|(val))
#define DP_REG16_CLR(reg, val)    DP_REG16_WRITE(reg,DP_REG16_READ(reg)&~(val))

/* Debug Macros */
#define DP_DEBUG_PROBE            0x00000001
#define DP_DEBUG_OPEN             0x00000002
#define DP_DEBUG_CLOSE            0x00000004
#define DP_DEBUG_IOCTL            0x00000008
#define DP_DEBUG_TX               0x00000010
#define DP_DEBUG_RX               0x00000020
#define DP_DEBUG_MC               0x00000040
#define DP_DEBUG_ANEG             0x00000080
#define DP_DEBUG_INTR             0x00000100
#define DP_DEBUG_LOAD             0x00000200
#define DP_DEBUG_UNLOAD           0x00000400

#if (DRV_DEBUG > 0)
static u32 dp_debug_level=DRV_DEBUG;
#define DP_DEBUG(level, X)  if (level & dp_debug_level) printk X
#else
#define DP_DEBUG(level, X)
#endif

/* data types */
typedef u32 status;			/* return status */
typedef volatile u8 *virt_addr;		/* CPU virtual address */
typedef volatile u8 *bus_addr;		/* BUS physical address */
typedef u8 bool;

#define OK	0			/* status: OK */
#define ERROR	-1			/* status: ERROR */
#define	TRUE	1
#define FALSE	0

/* Default Driver Parameters */
#define DP_DEFAULT_TXQ_SIZE		64
#define DP_DEFAULT_RXQ_SIZE		64
#define DP_RX_COPY_THRESHOLD		128 /* upper limit for rx packet copy */
#define DP_POLYNOMIAL			0x04C11DB6

/* Alignment and packet sizes */
#define ETH_CRC_LEN			4
#define ETH_MAX_PKT_SIZE		(ETH_FRAME_LEN + ETH_CRC_LEN)
#define DP_ALIGN			4     /* word alignment */

/* Driver private descriptor macros */
#define DP_DESC_SKBPTR			0x0c  /* SKB pointer offset */
#define DP_QUEUE_ELE_SIZE		(DP_DESC_SIZE + 4)
#define DP_QUEUE_ELE_NEXT(q) \
        q->cur_desc_addr = DP_QUEUE_ELE_NEXT_GET(q, q->cur_desc_addr)
#define DP_QUEUE_ELE_NEXT_GET(q, desc_addr) \
        ((desc_addr) == (q)->last_desc_addr) ? \
		(q)->first_desc_addr : (desc_addr) + DP_QUEUE_ELE_SIZE

/* Macros to get/set the values of the descriptor fields */
#define DP_DESC_LNK_GET(ptr)     *(u32 *)((virt_addr)ptr + DP_DESC_LNK)
#define DP_DESC_CMDSTS_GET(ptr)  *(u32 *)((virt_addr)ptr + DP_DESC_CMDSTS)
#define DP_DESC_BUFPTR_GET(ptr)  *(u32 *)((virt_addr)ptr + DP_DESC_BUFPTR)
#define DP_DESC_SKBPTR_GET(ptr)  *(u32 *)((virt_addr)ptr + DP_DESC_SKBPTR)

#define DP_DESC_LNK_SET(ptr, val)      DP_DESC_LNK_GET(ptr) = (val)
#define DP_DESC_CMDSTS_SET(ptr, val)   DP_DESC_CMDSTS_GET(ptr) = (val)
#define DP_DESC_BUFPTR_SET(ptr,val)    DP_DESC_BUFPTR_GET(ptr) = (val)
#define DP_DESC_SKBPTR_SET(ptr,val)    DP_DESC_SKBPTR_GET(ptr) = (val)

/*
 * Macros to get/set the values of descriptor fields with
 * appropriate address and byte order translations
 */
#define DP_DESC_LNK_XLATE_GET(p)    BUS_TO_CPU_ADDR_XLATE(DP_DESC_LNK_GET(p))
#define DP_DESC_CMDSTS_XLATE_GET(p) BUS_TO_CPU_SWAP_32(DP_DESC_CMDSTS_GET(p))
#define DP_DESC_BUFPTR_XLATE_GET(p) BUS_TO_CPU_ADDR_XLATE(DP_DESC_BUFPTR_GET(p))

#define DP_DESC_LNK_XLATE_SET(p, v) \
        DP_DESC_LNK_SET(p, CPU_TO_BUS_ADDR_XLATE(v))
#define DP_DESC_CMDSTS_XLATE_SET(p, v) \
        DP_DESC_CMDSTS_SET(p, CPU_TO_BUS_SWAP_32(v))
#define DP_DESC_BUFPTR_XLATE_SET(p,v) \
        DP_DESC_BUFPTR_SET(p, CPU_TO_BUS_ADDR_XLATE(v))

/* default things to check on phy interrupt */
#define DP_PHYSTS_LINK_MASK  (DP_PHYSTS_LNK_VALID | DP_PHYSTS_SPEED_10 | \
                              DP_PHYSTS_FDX)
				     
/* Descriptor queue */
struct dp83815_queue {
	virt_addr  first_desc_addr;  /* descriptor array address */
	virt_addr  last_desc_addr;   /* last descriptor address */
	virt_addr  cur_desc_addr;    /* current descriptor address */ 
	virt_addr  qbuf;             /* allocated queue buffer */
	u16        count;            /* number of elements */
};

/* Queue types -- qtype */
#define DP_QUEUE_TYPE_TX    1    /* Transmit queue */
#define DP_QUEUE_TYPE_RX    2    /* Receive queue */

/* Device private data */
struct dp83815_priv {
	struct device           *next;       /* Next dp83815 device */
	struct dp83815_queue    tx_queue;    /* Transmit Descriptor Queue */
	struct dp83815_queue    rx_queue;    /* Receive Descriptor Queue */
	struct enet_statistics  stats;       /* MIB data */
	char   hw_mac_addr[ETH_ALEN];	     /* Factory Ethernet Address */
	struct timer_list       timer;       /* timer for PHY setup */
	u16                     phy_status;  /* link status */
	u32                     phy_negotiate; /* link negotation */
};

static struct device *dp83815_dev_list;	/* List of dp83815 devices */

/* structure for hardware_id information of the device */
struct dp83815_hw_id {
	u16 vendor_id;
	u16 board_id;
};

/* 
 * Table for hardware_id information of the DP device. The first entry
 * to the table is the vendor_id and board_id for NS's Dp83815. Other
 * Vendors' extensions to the DP device should be added to this table.  
 */

static struct dp83815_hw_id dp83815_hw_id_tbl[] = {
        { PCI_VENDOR_ID_NS, PCI_DEVICE_ID_NS_83815 }
	/* add vendor extensions here */
};


/* Linux Network Driver interface routines */
extern int dp83815_probe(struct device *dev);
static void dp83815_link_setup(struct device *dev, const unsigned char);
static int dp83815_open(struct device *dev);
static int dp83815_close(struct device *dev);
static int dp83815_start_xmit(struct sk_buff *skb, struct device *dev);
static void dp83815_set_multicast_list(struct device *dev);
static int dp83815_ioctl(struct device *dev, struct ifreq *rq, int cmd);
static void dp83815_interrupt(int irq, void *dev_id, struct pt_regs *regs);
static struct enet_statistics *dp83815_get_stats(struct device *dev);

/* Support Functions */
static u16 swap_16(u16 us);
/*static u32 swap_32 (u32 ui);*/
static u32 io_read_32(u16 io_port);
static void io_write_32(u16 io_port, u32 data);
static u16 io_read_16(u16 io_port);
/*static void io_write_16 (u16 io_port, u16 data);*/

/* Driver Private Routines */
static int dp83815_pci_find_device(u16 pci_vendor_id, u16 pci_board_id, 
				   u8 *pbus, u8 *pfunc);
static void dp83815_mac_address_set(u32 iobase, char *mac_addr);
static void dp83815_mac_address_get(u32 iobase, char *mac_addr);
static status dp83815_dev_reset(u32 iobase);
static status dp83815_queue_create(struct dp83815_queue *q, int count, int qtype);
static status dp83815_queue_delete(struct dp83815_queue *q);
static virt_addr dp83815_tx_desc_get(struct device *dev);
static virt_addr dp83815_rx_desc_get(struct device *dev);
static status dp83815_phy_setup(struct device *dev);
static int dp83815_crc(char *mc_addr);

/* Driver Debug Routines */
#if (DRV_DEBUG > 0)
static void dp83815_regs_info(u16 iobase);
#endif

/* dp83815_tx_skb_reclaim - reclaim SKBs in transmitted descriptors */
static inline void
dp83815_tx_skb_reclaim(struct device *dev)
{
    struct sk_buff *skb;
    struct dp83815_queue *q;
    virt_addr desc_addr;
	
    /* Reclaim buffers from all descriptors we own. */
    q = &((struct dp83815_priv *)(dev->priv))->tx_queue;
    desc_addr = q->first_desc_addr;

    /* FIXME: we do a linear search of the TX queue, due to problems that got
     * us out of sync - this needs to be fixed --TPH 
     */
    do {
	skb = (struct sk_buff *)DP_DESC_SKBPTR_GET(desc_addr);
        if (!(DP_DESC_CMDSTS_XLATE_GET(desc_addr) & DP_DESC_CMDSTS_OWN)
	 && skb && !skb->list) {
	    DP_KFREE_SKB(skb);
	    DP_DESC_SKBPTR_SET(desc_addr, 0);
	}
	desc_addr = DP_QUEUE_ELE_NEXT_GET(q, desc_addr);
    } while (desc_addr != q->first_desc_addr);

#if 0
    // FIXME: Tim changed this
    /* Reclaim buffers from all descriptors we own. */
    q = &((struct dp83815_priv *)(dev->priv))->tx_queue;
    while (((DP_DESC_CMDSTS_XLATE_GET(desc_addr) & DP_DESC_CMDSTS_OWN) == 0) &&
	   ((skb=(struct sk_buff *) DP_DESC_SKBPTR_GET(desc_addr)) != NULL)) {
      DP_KFREE_SKB (skb); 
      DP_DESC_SKBPTR_SET (desc_addr, 0);
      desc_addr = DP_QUEUE_ELE_NEXT_GET (q, desc_addr);
    }
#endif
}


#ifdef DP_WOL_SUPPORT

/* for WOL support, we need to do the following:
 * 1) clear out the RXDP register,
 * 2) reset the receive logic
 * 3) clear any flags 
 * 4) set up receive logic */
static int dp83815_shutdown_notifier(struct notifier_block *this,
				     unsigned long code, void *x)
{
	struct device *dev = dp83815_dev_list;
	u16 iobase, i;
	
	if (!((code == SYS_HALT) || (code == SYS_POWER_OFF)))
		return NOTIFY_DONE;

	printk(KERN_INFO "setting up dp83815 for WOL\n");
	while (dev) {
		iobase = dev->base_addr;
		DP_REG32_WRITE(DP_RXDP, 0);
                /* reset and disable receiver */
		DP_REG32_WRITE(DP_CR, DP_CR_RXR | DP_CR_RXD); 
		DP_REG32_READ(DP_WCSR); /* read WCSR */
		for (i = 0; i < 100; i++) {
			/* read ISR until the reset is complete and then
			   enable WOL. */
			if (DP_REG32_READ(DP_ISR) & DP_ISR_RXRCMP) {
				DP_REG32_WRITE(DP_CCSR, DP_CCSR_PMESTS | DP_CCSR_PMEEN);
				DP_REG32_WRITE(DP_CR, DP_CR_RXE); 
				break;
			}
			udelay(5); /* wait a little bit */
		}
		dev = ((struct dp83815_priv *) dev->priv)->next;
	}
	return NOTIFY_DONE;
}

struct notifier_block dp83815_notifier = {
	dp83815_shutdown_notifier,
	NULL,
	0
};
#endif

/*
 * dp83815_probe - enumerate the PCI bus for instances of DP83815
 *
 * This routine enumerates DP83815 ethernet devices on the PCI bus, and
 * registers each DP83815 device found.
 */
int
dp83815_probe(struct device *dev)
{
  int i, j = 0;	
  int  board_count = (sizeof(dp83815_hw_id_tbl) / sizeof(struct dp83815_hw_id));
  u16  pci_vendor_id;
  u16  pci_board_id;
  int  dev_count, option;
  u8   bus;
  u8   func;
  u8   irq;
  u32  iobase;
  u32  version;
  char ether_addr[ETH_ALEN];
  struct dp83815_priv * priv;

  if (! pcibios_present())
    return -ENODEV;        /* No such device */

#ifdef DP_WOL_SUPPORT
  register_reboot_notifier(&dp83815_notifier);
#endif

  for (i=0, dev_count=0; i<board_count; i++) {
    /* get vendor id and board id from table */
    pci_vendor_id = dp83815_hw_id_tbl[i].vendor_id;
    pci_board_id = dp83815_hw_id_tbl[i].board_id;

    /* find the devices for each set of vendor and board id */
    for (dev = NULL ; 
    (dp83815_pci_find_device(pci_vendor_id, pci_board_id, &bus, &func) == TRUE);
    dev = NULL, j++) {
      /* Read PCI Configuration Registers */
      pcibios_read_config_byte(bus, func, PCI_INTERRUPT_LINE, &irq);
      pcibios_read_config_dword(bus, func, PCI_BASE_ADDRESS_0, &iobase);
      iobase &= PCI_BASE_ADDRESS_IO_MASK;
      option = i < MAX_UNITS ? options[dev_count] : 0;

      /* Put the device in a quiescent state */
      if (dp83815_dev_reset(iobase) != OK) {
	printk (KERN_INFO "%s: Device Reset failed -- bus=%d func=%d "
		"iobase=0x%x.\n",
		DP_DRV_NAME, bus, func, iobase);
	continue; /* Try the next device */
      }

      DP_REG32_WRITE(DP_CCSR, 0x0); /* XXX: make sure PME isn't enabled.
				       any WOL handling should go here. */

      /* Get the ethernet address */
      dp83815_mac_address_get(iobase, ether_addr);

      /* Allocate, name the device and add to ethernet device list */
      dev = init_etherdev(dev, sizeof(struct dp83815_priv));

      if (dev == NULL) {
	printk (KERN_INFO "%s: Failed to allocate device struct -- "
		"bus=%d func=%d iobase=0x%x\n",
		DP_DRV_NAME, bus, func, iobase);
	break;
      }
      priv=(struct dp83815_priv *)dev->priv;
            
      /* Save ethernet address */
      memcpy(dev->dev_addr, ether_addr, ETH_ALEN);
      memcpy(priv->hw_mac_addr, ether_addr, ETH_ALEN);

      /* If address is not set, set up a default one */
      /* FIXME: don't do this -- TPH */
      if ((dev->dev_addr[0] == 0) && (dev->dev_addr[1] == 0) &&
	  (dev->dev_addr[2] == 0) && (dev->dev_addr[3] == 0) &&
	  (dev->dev_addr[4] == 0) && (dev->dev_addr[5] == 0)) {
	u32 random = jiffies;
	u8 *ptr = (u8 *) &random;

	dev->dev_addr[0] = 0x08; /* National's Ethernet ID 0 */
	dev->dev_addr[1] = 0x00; /* National's Ethernet ID 1 */
	dev->dev_addr[2] = 0x17; /* National's Ethernet ID 2 */
	dev->dev_addr[3] = *ptr++;
	dev->dev_addr[4] = *ptr++;
	dev->dev_addr[5] = *ptr + j;
	printk("%s: no mac address (nomacaddr)\n", dev->name);
      } else if ((dev->dev_addr[0] == 0x00) && (dev->dev_addr[1] == 0x08)){
	/* Buggy ethernet address; needs swapping */
	/* FIXME: wtf is this about? -- TPH */
	u16 *ptr = (u16 *) dev->dev_addr;
	u8 i;
			
	for (i=0; i<3; i++) {
	  *ptr = swap_16(*ptr);
	  ptr++;
	}
      }

      dp83815_mac_address_set(iobase, dev->dev_addr);
		
      /* initialize the device data */
      dev->base_addr = iobase;
      dev->irq       = irq;
      dev->open      = dp83815_open;
      dev->stop      = dp83815_close;
      dev->get_stats = dp83815_get_stats;
      dev->do_ioctl  = dp83815_ioctl;
      dev->hard_start_xmit    = dp83815_start_xmit;
      dev->set_multicast_list = dp83815_set_multicast_list;

      ether_setup (dev);        /* initialize generic fields */
		
      /* reserve IO region */
      request_region (iobase, DP83815_PCI_IO_SIZE, dev->name);
		
      /* Display board info */
      version = DP_REG32_READ(DP_SRR);
      printk (KERN_INFO "%s: National Semiconductor MacPhyter (dp83815)\n",
		dev->name);

      printk (KERN_INFO "%s:   bus=%d func=%d io=0x%x irq=%d ver=%d.%d\n",
	      dev->name, bus, func, iobase, irq,
	      (version & DP_SRR_MAJ) >> DP_SRR_MAJ_SHF,
	      (version & DP_SRR_MIN));
		
      printk (KERN_INFO "%s:   ethernet addr=%02x:%02x:%02x:%02x:%02x:%02x\n",
	      dev->name,
	      dev->dev_addr[0], dev->dev_addr[1], dev->dev_addr[2],
	      dev->dev_addr[3], dev->dev_addr[4], dev->dev_addr[5]);
		
      if (dev->mem_start)
	      option = dev->mem_start;
      
      /* The lower four bits are the media type. */
      priv->phy_negotiate = DP_CFG_ANEG_SEL_ALL_XD;
      if (((option > 0) && (option & 0x200)) ||
	  (dev_count < MAX_UNITS && full_duplex[dev_count] > 0))
	      priv->phy_negotiate = DP_CFG_ANEG_SEL_100_FD;

      /* Chain the device */
      ((struct dp83815_priv *)(dev->priv))->next = dp83815_dev_list;
      dp83815_dev_list = dev;
		
#ifdef DEBUG
      printk (KERN_INFO "%s: DebugInfo: dev=0x%x priv=0x%x "
	      "&dp_debug_level=0x%x\n",
	      dev->name, (u32)dev, (u32)dev->priv, (u32)&dp_debug_level);
#endif

      /* Update Counters */
      dev_count++;
    }
  }

#ifdef CONFIG_COBALT_GEN_V
  if (dev_count)
	  lcd_handler_add(&dp83815_lcd_handler);
#endif

  return dev_count ? OK : -ENODEV;
}

/* setup a timer to deal with link changes */
static void dp83815_timer(unsigned long data)
{
	struct device *dev = (struct device *) data;
	struct dp83815_priv *priv = (struct dp83815_priv *) dev->priv;
	u32 iobase = dev->base_addr, page, dspcfg;
	unsigned long flags;

	page = DP_REG32_READ(DP_PHY_PAGE);
        DP_REG32_WRITE (DP_PHY_PAGE, DP_PHY_PAGE_VAL);
	dspcfg = DP_REG32_READ(DP_PHY_DSPCFG);
	DP_REG32_WRITE(DP_PHY_PAGE, page);
	if (dspcfg != DP_PHY_DSPCFG_VAL) {	
		save_flags(flags);
		cli();
		if (!dev->tbusy) { /* phy setup clears things */
			printk(KERN_INFO "%s: possible phy reset: re-initializing\n", 
		       	       dev->name);
			dp83815_phy_setup(dev);
		}
		restore_flags(flags);
	} else
		dp83815_link_setup(dev, 0);
	priv->timer.expires = jiffies + HZ;
	add_timer(&priv->timer);
}


/* dp83815_open - open and initialize a device */
static int
dp83815_open (struct device *dev)
{
  u32 iobase = dev->base_addr;
  struct dp83815_priv *priv = dev->priv;
 
  /* Allocate Tx and Rx queues */
  if (dp83815_queue_create (&priv->tx_queue, DP_DEFAULT_TXQ_SIZE,
    DP_QUEUE_TYPE_TX) != OK) {
    printk (KERN_INFO "%s: Failed to create tx queue\n", dev->name);
    return -EAGAIN;
  }

  if (dp83815_queue_create (&priv->rx_queue, DP_DEFAULT_RXQ_SIZE,
    DP_QUEUE_TYPE_RX) != OK) {
    dp83815_queue_delete (&priv->tx_queue);
    printk (KERN_INFO "%s: Failed to create rx queue\n", dev->name);
    return -EAGAIN;
  }

  /* Install the Tx and Rx queues on the device */
  DP_REG32_WRITE (DP_TXDP, virt_to_bus(priv->tx_queue.first_desc_addr));
  DP_REG32_WRITE (DP_RXDP, virt_to_bus(priv->rx_queue.first_desc_addr));

  DP_DEBUG (DP_DEBUG_OPEN,
	    (KERN_INFO "dp: setting TXDP=0x%x RXDP=0x%x\n",
	     (u32)virt_to_bus (priv->tx_queue.first_desc_addr),
	     (u32)virt_to_bus (priv->rx_queue.first_desc_addr)));

  /* Install interrupt vector */
  if (request_irq (dev->irq, &dp83815_interrupt, SA_SHIRQ,
		   dev->name, dev) != OK) {
    dp83815_queue_delete (&priv->tx_queue);
    dp83815_queue_delete (&priv->rx_queue);
    return -EAGAIN;
  }

  /* set link negotiation flags */

  /* Setup phy capabilities */
  if (dp83815_phy_setup (dev) != OK) 
    printk (KERN_INFO "%s: Warning PHY setup did not complete. Check cable.\n",
	    dev->name);

  /* Setup the ethernet address */
  dp83815_mac_address_set (iobase, dev->dev_addr);

  /* Receive perfect match and broadcast packets */
  DP_REG32_WRITE (DP_RFCR, 0);
  DP_REG32_WRITE (DP_RFCR, (DP_RFCR_AAB  | /* all broadcast pkts */
			    DP_RFCR_APM  | /* perfect match pkts */
			    DP_RFCR_RFEN));

  DP_DEBUG (DP_DEBUG_OPEN,
	    (KERN_INFO "dp: RFCR set to 0x%x\n",
	     DP_RFCR_RFEN | DP_RFCR_APM | DP_RFCR_AAB));
                    
  /* Turn on device interrupts */
  DP_REG32_WRITE (DP_IMR, (DP_IMR_RXOK  | DP_IMR_MIB   |
			   DP_IMR_RTABT | DP_IMR_RMABT |
			   DP_IMR_SSERR | DP_IMR_PHY));
  DP_REG32_WRITE (DP_MICR, DP_MICR_INTEN);
  DP_REG32_WRITE (DP_IER, DP_IER_IE);

  /* Enable Tx/Rx */
  DP_REG32_WRITE (DP_CR, DP_CR_TXE | DP_CR_RXE);

  /* Increment module reference count */
  MOD_INC_USE_COUNT;

  /* setup link heartbeat */
  init_timer(&priv->timer);
  priv->timer.expires = jiffies + 3*HZ;
  priv->timer.data = (unsigned long) dev;
  priv->timer.function = &dp83815_timer;
  add_timer(&priv->timer);

  return OK;
}

/* dp83815_close - close a device, and reclaim resources */
static int
dp83815_close(struct device *dev)
{
    u32 iobase = dev->base_addr;
    struct dp83815_priv *priv = dev->priv;

    /* Stop the Tx/Rx */
    /* Do we need to do this ??? */
    DP_REG32_WRITE (DP_CR, (DP_CR_TXD | DP_CR_RXD)); 

    /* Restore factory ethernet address */
    dp83815_mac_address_set (iobase, priv->hw_mac_addr);
    
    /* get rid of timer */
    del_timer(&priv->timer);

    /* turn off interrupt generation */
    DP_REG32_WRITE(DP_IER, 0);

    /* Uninstall the interrupt vector */
    free_irq (dev->irq, dev);

    /* Free the Tx and Rx queues */
    dp83815_queue_delete (&priv->tx_queue);
    dp83815_queue_delete (&priv->rx_queue);

    /* Decrement module reference count */
    MOD_DEC_USE_COUNT;
    
    return OK;
}

/*
 * dp83815_start_xmit - transmit an ethernet packet.
 *
 * This routine writes to a tx descriptor, sets the ownership bit of the
 * CMDSTS, and signals the chip
 */
static int
dp83815_start_xmit(struct sk_buff *skb, struct device *dev)
{
    u32 iobase = dev->base_addr;
    u32 cmdsts;
    virt_addr tx_desc;
    struct enet_statistics *stats_p;

    if (skb->len > ETH_MAX_PKT_SIZE) {
        return ERROR;
    }

    /* see if the current tx descriptor is available */
    tx_desc = dp83815_tx_desc_get(dev);
    if (tx_desc == NULL) {
	/* turn on interrupts on tx complete, and on tx idle */
        set_bit(0, &dev->tbusy);
        DP_REG32_SET(DP_IMR, DP_IMR_TXOK | DP_IMR_TXIDLE);
        tx_desc = dp83815_tx_desc_get(dev);
    }

    stats_p = &((struct dp83815_priv *)(dev->priv))->stats;

    /* Update tx_desc to point to SKB data, set CMDSTS, and signal the chip */
    if (tx_desc != NULL) {
        static unsigned int npkts = 0;
	u32 newsts = DP_DESC_CMDSTS_OWN;
	
        /* get info from the previous transmit */
        cmdsts = DP_DESC_CMDSTS_XLATE_GET(tx_desc);

	/* interrupt every other packet */
        if (!(npkts++ & 0x1)) {
            newsts |= DP_DESC_CMDSTS_INTR;
        }

        /* Update the descriptor */
        DP_DESC_BUFPTR_XLATE_SET(tx_desc, skb->data);
        DP_DESC_SKBPTR_SET(tx_desc, (u32)skb);
        DP_DESC_CMDSTS_XLATE_SET(tx_desc, newsts|skb->len);

	LED_SET_TXRX(dev);

        dev->trans_start = jiffies;
        DP_REG32_SET(DP_CR, DP_CR_TXE);
        DP_DEBUG(DP_DEBUG_TX, 
            (KERN_INFO "Tx: tx_desc=0x%x ", (u32)tx_desc));

        /* update statistics of the previous transmit */
        if (cmdsts & DP_DESC_CMDSTS_TX_ERRORS) {
	    /* general errors */
            stats_p->tx_errors++;

	    /* collisions */
            stats_p->collisions += DP_DESC_CMDSTS_TX_COLLISIONS_GET(cmdsts);

	    /* tx aborted */
            if (cmdsts & DP_DESC_CMDSTS_TX_TXA) {
                stats_p->tx_packets--;
            }

            /* fifo errors */	    
            if (cmdsts & DP_DESC_CMDSTS_TX_TFU) {
                stats_p->tx_fifo_errors++;
            }

	    /* lost carrier */
            if (cmdsts & DP_DESC_CMDSTS_TX_CRS) {
                stats_p->tx_carrier_errors++;
            }

	    /* out of window collisions */
            if (cmdsts & DP_DESC_CMDSTS_TX_OWC) {
                stats_p->tx_window_errors++;
            }
        }

#if 0 /* FIXME: should update TXDP here! */
        /* check for errors */
        cmdsts = DP_DESC_CMDSTS_XLATE_GET(tx_desc);
        if (!(cmdsts & DP_DESC_CMDSTS_OWN) 
         && (cmdsts & DP_DESC_CMDSTS_TX_ERRORS)) {
            /* error on this xmit - try again */    
            DP_DESC_BUFPTR_XLATE_SET(tx_desc, skb->data);
            DP_DESC_SKBPTR_SET(tx_desc, (u32)skb);
            DP_DESC_CMDSTS_XLATE_SET(tx_desc, newsts|skb->len);
            dev->trans_start = jiffies;
            DP_REG32_SET(DP_CR, DP_CR_TXE);
            DP_DEBUG(DP_DEBUG_TX, 
                (KERN_INFO "Tx-retry: tx_desc=0x%x ", (u32)tx_desc));
        }
#endif
       
        DP_STATS_INCR(stats_p, tx, skb->len);
    } else {
	return 1;
    }

    return OK;
}

#ifdef CONFIG_COBALT_GEN_III
/*
 * Under very high load we don't always get to the watchdog
 * refresh routine.  Re-arm it here just to be safe.
 */             
#include <linux/cobalt_wdt.h>

#define COBALT_WDT_CLEARDOG cobalt_wdt_cleardog()
#else   
#define COBALT_WDT_CLEARDOG
#endif 


/*
 * dp83815_start_receive - receive the data from a Rx Queue descriptor
 *
 * This routine receives the data from Rx queue as long as it gets a valid
 * rx_descriptor and resets the descriptor's CMDSTS field back to the Buffer
 * size, and updates the BUFPTR and SKBPTR fields to the newly allocated SKB.
 */
static int
dp83815_start_receive (struct device *dev)
{
    u32	iobase = dev->base_addr;
    u32	cmdsts;
    int len;
    bool do_copy;
    virt_addr rx_desc;
    struct sk_buff *cur_skb;
    struct sk_buff *new_skb;
    struct sk_buff *rx_skb;
    struct enet_statistics *stats_p;
    int npkts = 0;
    
    stats_p = &((struct dp83815_priv *)(dev->priv))->stats;
    
    LED_SET_TXRX(dev);
    for (rx_desc = dp83815_rx_desc_get(dev);
         (rx_desc != NULL);
		 rx_desc = dp83815_rx_desc_get(dev)) {
        
        DP_DEBUG (DP_DEBUG_RX, 
                  (KERN_INFO "Rx: rx_desc=0x%x, CMDSTS = 0x%x",
                   (u32)rx_desc, DP_DESC_CMDSTS_XLATE_GET(rx_desc)));

	if (++npkts > DP_DEFAULT_RXQ_SIZE / 2) {
	    npkts = 0;
	    COBALT_WDT_CLEARDOG;
	}
        cmdsts = DP_DESC_CMDSTS_XLATE_GET (rx_desc);

        /* Send the packet to the stack if no errors */
        if ((cmdsts & DP_DESC_CMDSTS_RX_ERRORS) == 0
	 && (cmdsts & DP_DESC_CMDSTS_SIZE)) {
            len = (cmdsts & DP_DESC_CMDSTS_SIZE) - ETH_CRC_LEN;
            /*
             * Allocate a new SKB
             * small data packets less than DP_RX_COPY_THRESHOLD are copied
             * into the new SKB, other allocate one to replace the current SKB.
             */
            if (len < DP_RX_COPY_THRESHOLD) {
                do_copy = TRUE;
                new_skb = alloc_skb(len + 2, GFP_ATOMIC);
            } else {
                do_copy = FALSE;
                new_skb = alloc_skb(ETH_MAX_PKT_SIZE, GFP_ATOMIC);
            }
		
            if (new_skb) {
                cur_skb = (struct sk_buff *)DP_DESC_SKBPTR_GET(rx_desc);
		if (do_copy) {
			/* Copy data from current SKB and send the new SKB up */
			rx_skb = new_skb;
			skb_reserve(rx_skb, 2);
			if (rx_skb->tail+len <= rx_skb->end) {
			    memcpy(skb_put(rx_skb, len), cur_skb->data, len);
			} else {
			    /* FIXME: */
                	    stats_p->rx_dropped++; /* no resources */
			}
		} else {
			/* Replace the the current SKB with the new SKB */
			rx_skb = cur_skb;
			if (rx_skb->tail+len <= rx_skb->end) {
			    DP_DESC_BUFPTR_XLATE_SET(rx_desc, new_skb->data);
			    DP_DESC_SKBPTR_SET(rx_desc, (u32)new_skb);
			    (void)skb_put(rx_skb, len);
			} else {
			    /* FIXME: */
                	    stats_p->rx_dropped++; /* no resources */
			}
		}

		/* update the SKB and set it up */
		rx_skb->dev = dev;
		rx_skb->protocol = eth_type_trans(rx_skb, dev);

		netif_rx(rx_skb);

		dev->last_rx = jiffies;
		DP_STATS_INCR(stats_p, rx, len);
            } else {
                stats_p->rx_dropped++; /* no resources */
	    }
	    if (cmdsts & DP_DESC_CMDSTS_RX_DEST_MC) {
		stats_p->multicast++;
	    }
        } else if (cmdsts & DP_DESC_CMDSTS_RX_ERRORS) {
            stats_p->rx_errors++; /* bad packet */

            /* Update individual counters */
            if (cmdsts & (DP_DESC_CMDSTS_RX_RUNT | DP_DESC_CMDSTS_RX_LONG))
                stats_p->rx_length_errors++;

            if (cmdsts & DP_DESC_CMDSTS_RX_CRCE)
                stats_p->rx_crc_errors++;

            if (cmdsts & DP_DESC_CMDSTS_RX_FAE)
                stats_p->rx_frame_errors++;

            if (cmdsts & DP_DESC_CMDSTS_RX_RXO)
                stats_p->rx_fifo_errors++;
	} else {
		printk("%s: odd, len is 0 (cmdsts: %x)\n", dev->name, cmdsts);
	}

        /* Cleanup the descriptor and make available for reception */
        DP_DESC_CMDSTS_XLATE_SET (rx_desc, ETH_MAX_PKT_SIZE);
	DP_REG32_SET (DP_CR, DP_CR_RXE);
    }
    return OK;
}

/* dp83815_get_stats - get current device statistics */
static struct enet_statistics *
dp83815_get_stats (struct device *dev)
{
    return &((struct dp83815_priv *)(dev->priv))->stats;
}

/* dp83815_set_multicast_list - sets multicast, & promiscuous mode */
static void
dp83815_set_multicast_list (struct device *dev)
{
    u32 iobase = dev->base_addr;
    u16 hash_table[32];
    u32 rfcr_flags;
    int i;
    struct dev_mc_list *mc_list;

    /* default RFCR mode */
    rfcr_flags = DP_RFCR_APM | DP_RFCR_AAB | DP_RFCR_RFEN;

    /* Setup promiscuous mode */
    if (dev->flags & IFF_PROMISC) {
        DP_DEBUG (DP_DEBUG_OPEN,
                  (KERN_INFO "IFF_PROMISC\n"));
        rfcr_flags = (DP_RFCR_AAU | DP_RFCR_AAM | DP_RFCR_AAB |
                      DP_RFCR_RFEN);
    } else if (dev->flags & IFF_ALLMULTI) {
        /* Receive all multicast packets */
        DP_DEBUG (DP_DEBUG_OPEN,
                  (KERN_INFO "IFF_ALLMULTI\n"));
        rfcr_flags |= DP_RFCR_AAM;
    } else {
        /* Setup to receive programmed multicast packets */
        memset (hash_table, 0, 32);
        for (i=0, mc_list=dev->mc_list; mc_list && i < dev->mc_count;
             i++, mc_list = mc_list->next) {
		DP_DEBUG (DP_DEBUG_OPEN,
			  (KERN_INFO "mc_addr=%p\n", mc_list->dmi_addr));
            set_bit(dp83815_crc((char *)mc_list->dmi_addr) & 0x1ff, hash_table);
        }

        /* install the hash table */
        for (i=0; i<32; i++) {
            DP_REG32_WRITE (DP_RFCR, DP_RFCR_RFADDR_FMEM_LO + i*2);
            DP_REG32_WRITE (DP_RFDR, (u32) hash_table[i]);
        }

        rfcr_flags |= DP_RFCR_MHEN;
    }

    DP_REG32_WRITE (DP_RFCR, 0);
    DP_REG32_WRITE (DP_RFCR, rfcr_flags);
    DP_DEBUG (DP_DEBUG_OPEN,
              (KERN_INFO "MC Setup RFCR flags=0x%x\n", rfcr_flags));

    return;
}

/* dp83815_crc - computer CRC for hash table entries */
static int
dp83815_crc (char * mc_addr)
{
    u32 crc;
    u8 cur_byte;
    u8 msb;
    u8 byte, bit;

    crc = ~0;
    for (byte=0; byte<6; byte++) {
        cur_byte = *mc_addr++;
        for (bit=0; bit<8; bit++) {
            msb = crc >> 31;
            crc <<= 1;
            if (msb ^ (cur_byte & 1)) {
                crc ^= DP_POLYNOMIAL;
                crc |= 1;
            }
            cur_byte >>= 1;
        }
    }
    crc >>= 23;

    return (crc);
}

/* dp83815_ioctl - handle driver specific ioctls */
static int
dp83815_ioctl(struct device *dev, struct ifreq *rq, int cmd)
{
    return -EOPNOTSUPP;
}

/* dp83815_interrupt - handle the interrupts */
static void
dp83815_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    struct device *dev = dev_id;
    u16 iobase = dev->base_addr;
    u32 reg_isr;

    reg_isr = DP_REG32_READ(DP_ISR);
    
    DP_DEBUG(DP_DEBUG_IOCTL,
              (KERN_INFO "%s: intr_status=0x%x\n", dev->name, reg_isr));

    /* check for a phy interrupt */
    if (reg_isr & DP_ISR_PHY) 
	    dp83815_link_setup(dev, 0);

    if (reg_isr & DP_ISR_RXOK) {
        dp83815_start_receive(dev);
    }

    if (reg_isr & (DP_ISR_TXOK | DP_ISR_TXIDLE)) {
        DP_REG32_CLR (DP_IMR, (DP_IMR_TXOK|DP_IMR_TXIDLE));
        clear_bit (0, &dev->tbusy);
        mark_bh (NET_BH);
    }

    /* Reclaim SKBs from the TX queue */
    dp83815_tx_skb_reclaim(dev);
}

/* dp83815_mac_address_set - set the ethernet address */
static void
dp83815_mac_address_set(u32 iobase, char *mac_addr)
{
    u16 *mac_addr_ptr;
    int i;

    for (i=0, mac_addr_ptr = (u16 *)mac_addr; i<3; i++, mac_addr_ptr++) {
        DP_REG32_WRITE (DP_RFCR, DP_RFCR_RFADDR_PMATCH1 + i*2);
        DP_REG32_WRITE (DP_RFDR, CPU_TO_BUS_SWAP_16 (*mac_addr_ptr));
    }
}

/* dp83815_pci_find_device - call the kernel-version appropriate routine to
 * find the pci devices
 */
static int 
dp83815_pci_find_device (u16 pci_vendor_id, u16 pci_board_id,
			 u8 *pbus, u8 *pfunc)
{
#ifdef DP_PCIBIOS_FIND_DEVICE
  if (pcibios_find_device (pci_vendor_id, pci_board_id, dp_dev_cur,
			   pbus, pfunc) == PCIBIOS_SUCCESSFUL) {
    dp_dev_cur++;
    return TRUE;
  }
  else
    return FALSE;
#endif
#ifdef DP_PCI_FIND_DEVICE
  struct pci_dev *newdev;
  /* we use newdev because we want to never make dp_pci_dev = NULL */
  newdev = pci_find_device (pci_vendor_id, pci_board_id, dp_pci_dev);
  if (newdev) {
    dp_pci_dev = newdev;
    *pbus = dp_pci_dev->bus->number;
    *pfunc = dp_pci_dev->devfn;
    return TRUE;
  }
  else
    return FALSE;
#endif
}

/* dp83815_mac_address_get - get the ethernet address */
static void
dp83815_mac_address_get(u32 iobase, char *mac_addr)
{
    u16 *mac_addr_ptr;
    int    i;

    for (i=0, mac_addr_ptr = (u16*)mac_addr; i<3; i++, mac_addr_ptr++) {
        DP_REG32_WRITE (DP_RFCR, DP_RFCR_RFADDR_PMATCH1 + i*2);
        *mac_addr_ptr = BUS_TO_CPU_SWAP_16 (DP_REG32_READ (DP_RFDR));
    }
}

/* dp83815_dev_reset - soft reset DP83815 */
static status
dp83815_dev_reset(u32 iobase)
{
    int  timeout;
    
    /* reload eeprom */
    DP_REG32_WRITE(DP_PTSCR, DP_PTSCR_EELOAD_EN);
    timeout = 400; /* need to wait > 1500us */
    while (timeout--) {
	if ((DP_REG32_READ(DP_PTSCR) & DP_PTSCR_EELOAD_EN) == 0) 
            return (OK);
	udelay(5);
    }

    /* reset mac */
    DP_REG32_WRITE(DP_CR, DP_CR_RST);
    timeout = 100;
    while (timeout--) {
        if (DP_REG32_READ(DP_ISR) == 0x03008000) {
            return (OK);
        }
        udelay(5);
    }
    return ERROR;
}

/*
 * dp83815_queue_create - create a circular queue of descriptors
 *
 * This routine allocates a descriptor buffer array aligned  on a word
 * boundary, initializes and links the array to make a circular queue.
 */
static status
dp83815_queue_create (struct dp83815_queue *q, int count, int qtype)
{
    virt_addr desc_addr;
    int i;
    struct sk_buff *skb;
    
    /* allocate the desc buffer array */
    q->qbuf = (virt_addr) kmalloc (DP_QUEUE_ELE_SIZE * count + DP_ALIGN,
                                   GFP_DMA);
    if (q->qbuf == (virt_addr) NULL)
        return ERROR;

    memset ((char *)q->qbuf, 0, DP_QUEUE_ELE_SIZE * count + DP_ALIGN);

    /* Adjust alignment and Initialize queue data */
    q->cur_desc_addr = q->first_desc_addr =
        (virt_addr)(((u32)q->qbuf + DP_ALIGN) & ~(DP_ALIGN - 1));
    q->last_desc_addr = q->first_desc_addr + ((count -1) * DP_QUEUE_ELE_SIZE);
    q->count = count;

    /* Initialize each buffer descriptor, and link them into circular queue */
    for (i=0, desc_addr=q->first_desc_addr; i<count;
         i++, desc_addr+=DP_QUEUE_ELE_SIZE) {
        DP_DESC_LNK_XLATE_SET (desc_addr, desc_addr + DP_QUEUE_ELE_SIZE);

        /* Update the size, BUFPTR, and SKBPTR fields for RX descriptors */
        if (qtype == DP_QUEUE_TYPE_RX) {
            skb = alloc_skb (ETH_MAX_PKT_SIZE, GFP_ATOMIC);
            if (skb == NULL) {
                dp83815_queue_delete (q);
                return (ERROR);
            }
            DP_DESC_CMDSTS_XLATE_SET (desc_addr, ETH_MAX_PKT_SIZE);
            DP_DESC_BUFPTR_XLATE_SET (desc_addr, skb->data);
            DP_DESC_SKBPTR_SET (desc_addr, (u32) skb);
        }
    }
    /* Make the queue circular */
    DP_DESC_LNK_XLATE_SET (q->last_desc_addr, q->first_desc_addr);

    return OK;
}

/* dp83815_queue_delete - frees an allocated descriptor queue */
static status
dp83815_queue_delete (struct dp83815_queue *q)
{
    int       i;
    virt_addr desc_addr;
    struct sk_buff *skb;
        
    /* Free all SKBs in the queue */
    for (i=0, desc_addr=q->first_desc_addr;
         (i < q->count);
         i++, desc_addr += DP_QUEUE_ELE_SIZE) {
        skb = (struct sk_buff *) DP_DESC_SKBPTR_GET (desc_addr);
        if (skb != NULL)
	  DP_KFREE_SKB (skb); 
    }

    /* Free the queue buffer */
    kfree ((char *)q->qbuf);
    
    return (OK);
}

/*
 * dp83815_tx_desc_get - get a valid transmit descriptor
 *
 * This routine returns the current descriptor from the tx_queue if driver is 
 * the owner, else returns NULL
 */
static virt_addr
dp83815_tx_desc_get(struct device *dev)
{
    struct dp83815_queue *q;
    virt_addr desc_addr = NULL;

    q = &((struct dp83815_priv *)(dev->priv))->tx_queue;

    /* Check if we own the descriptor, and it does not have an skb */
    if (!(DP_DESC_CMDSTS_XLATE_GET(q->cur_desc_addr) & DP_DESC_CMDSTS_OWN)
     && !(DP_DESC_SKBPTR_GET(q->cur_desc_addr))) {
        desc_addr = q->cur_desc_addr;
        /* Move to the next element */
        DP_QUEUE_ELE_NEXT (q); 
    }

    return desc_addr;
}

/*
 * dp83815_rx_desc_get - get a valid receive descriptor
 *
 * This routine returns the current descriptor from the rx_queue if driver is 
 * the owner, else returns NULL
 */
static virt_addr
dp83815_rx_desc_get (struct device *dev)
{
    struct dp83815_queue *q;
    virt_addr desc_addr = NULL;

    q = &((struct dp83815_priv *)(dev->priv))->rx_queue;

    /* Check if we own the descriptor */
    if  (DP_DESC_CMDSTS_XLATE_GET (q->cur_desc_addr) & DP_DESC_CMDSTS_OWN) {
        desc_addr = q->cur_desc_addr;
        DP_QUEUE_ELE_NEXT(q); /* Move to the next element */
    }

    return desc_addr;
}

/* check full-duplex */
static void dp83815_link_setup(struct device *dev, const unsigned char init)
{
    u32  iobase = dev->base_addr;
    struct dp83815_priv *priv = dev->priv;
    u16 phy_status = DP_REG16_READ (DP_PHYSTS) & DP_PHYSTS_LINK_MASK;
    int fdx, oldfdx;

    if (!init && (phy_status == priv->phy_status))
	    return;

    fdx = phy_status & DP_PHYSTS_FDX;
    oldfdx = priv->phy_status & DP_PHYSTS_FDX;
    priv->phy_status = phy_status;

    /* see if we're still auto-negotiating */
    if (DP_REG32_READ(DP_CFG) & DP_CFG_ANEG_DN) {
	    printk (KERN_INFO "%s: speed=%d duplex=%s link=%s\n", dev->name,
		    (phy_status & DP_PHYSTS_SPEED_10) ? 10 : 100,
		    fdx ? "full" : "half",
		    (phy_status & DP_PHYSTS_LNK_VALID) ? "up" : "down");
    } else if (init) 
	    printk(KERN_INFO "%s: auto-negotiating\n", dev->name);
    else
	    return;

    if (!init && (fdx == oldfdx))
	    return;

    if (fdx) {
	    /* Setup transmit control */
	    DP_REG32_WRITE (DP_TXCFG, (DP_TXCFG_DRTH_SET(0x30) | 
				       DP_TXCFG_FLTH_SET(0x10) |
				       DP_TXCFG_MXDMA_32       |
				       DP_TXCFG_ATP            |
				       DP_TXCFG_CSI            |
				       DP_TXCFG_HBI));

	    DP_DEBUG (DP_DEBUG_OPEN,
		      (KERN_INFO "dp (Full Duplex): TXCFG set to 0x%x\n",
		       (DP_TXCFG_DRTH_SET(0x30) | DP_TXCFG_FLTH_SET(0x10) |
			DP_TXCFG_MXDMA_32       | DP_TXCFG_ATP            |
			DP_TXCFG_CSI            | DP_TXCFG_HBI)));

	    /* Setup receive control */
	    DP_REG32_WRITE (DP_RXCFG, (DP_RXCFG_DRTH_SET(0x08) |
				       DP_RXCFG_MXDMA_32       |
				       DP_RXCFG_ATX));
		    
	    DP_DEBUG (DP_DEBUG_OPEN,
		      (KERN_INFO "dp (Full Duplex): RXCFG set to 0x%x\n",
		       (DP_RXCFG_DRTH_SET(0x08) | DP_RXCFG_MXDMA_32 |
			DP_RXCFG_ATX)));
    } else {
	    /* Setup transmit control */
	    DP_REG32_WRITE (DP_TXCFG, (DP_TXCFG_DRTH_SET(0x30) | 
				       DP_TXCFG_FLTH_SET(0x10) |
				       DP_TXCFG_MXDMA_32       |
				       DP_TXCFG_ATP));

	    DP_DEBUG (DP_DEBUG_OPEN,
		      (KERN_INFO "dp (Half Duplex): TXCFG set to 0x%x\n",
		       (DP_TXCFG_DRTH_SET(0x30) | DP_TXCFG_FLTH_SET(0x10) |
			DP_TXCFG_MXDMA_32       | DP_TXCFG_ATP)));
	    
	    /* Setup receive control */
	    DP_REG32_WRITE (DP_RXCFG, (DP_RXCFG_DRTH_SET(0x08) |
			       DP_RXCFG_MXDMA_32));

	    DP_DEBUG (DP_DEBUG_OPEN,
	      (KERN_INFO "dp (Half Duplex): RXCFG set to 0x%x\n",
	       (DP_RXCFG_DRTH_SET(0x08) | DP_RXCFG_MXDMA_32)));
    }
}

/* dp83815_phy_setup - reset and setup the PHY device */
static status
dp83815_phy_setup (struct device *dev)
{
    u32  iobase = dev->base_addr;
    struct dp83815_priv *priv = dev->priv;
    u32  dp_cfg_val;
    u16  timeout;

    dp_cfg_val =  (DP_CFG_PESEL           |     /* parity error detect */
                   priv->phy_negotiate    |     /* negotiate link */
                   DP_CFG_PAUSE_ADV       |     /* pause capable */
                   DP_CFG_PINT_ACEN       |     /* phy intr auto clear */
                   0x00040000);                 /* phy config */

    DP_REG32_WRITE (DP_CFG, dp_cfg_val | DP_CFG_PHY_RST);
    udelay (500);

    DP_REG32_WRITE (DP_CFG, dp_cfg_val);
    for (timeout=4; timeout; timeout--) {
        if (DP_REG32_READ (DP_CFG) & DP_CFG_ANEG_DN)
            break;
	mdelay(1);
    }

    /* if we don't set DSPCFG, then we can run into auto-negotiation
     * problems */
    DP_REG32_WRITE (DP_PHY_PAGE, DP_PHY_PAGE_VAL);
    DP_REG32_WRITE (DP_PHYCR, DP_PHYCR_PMDCSR_VAL);
    DP_REG32_WRITE (DP_PHY_TDATA, DP_PHY_TDATA_VAL);
    DP_REG32_WRITE (DP_PHY_DSPCFG, DP_PHY_DSPCFG_VAL);
    DP_REG32_WRITE (DP_PHY_SDCFG, DP_PHY_SDCFG_VAL);
    DP_REG32_WRITE (DP_PHY_PAGE, 0x0);

    /* setup transmit and receive control according to duplexity of Phy */
    dp83815_link_setup(dev, 1);
 
#if 0
    if (timeout == 0) {
        DP_DEBUG (DP_DEBUG_ANEG, (KERN_INFO "Phy Autonegotiate Failed\n"));
        return (ERROR);
    }
#endif

    return (OK);
}

#if 1
/* swap_16 - swap a 16 bit value between little endian & big endian byteorder */
static u16 swap_16 (u16 us)
{
    return ((us << 8) & 0xff00) | ((us >> 8) & 0x00ff);
}
#endif

#if 0 /* Not used */
/* swap_32 - swap a 32 bit value between little endian & big endian byteorder */
static u32 swap_32 (u32 ui)
{
    return ((swap_16 (ui & 0x0000ffff) << 16) | swap_16 (ui >> 16));
}
#endif

/* io_read_16 - read the 16 bit value stored at given io address */
static u16 io_read_16 (u16 io_port)
{
  return DP_INW(io_port); 
}

#if 0 /* Not used */
/* io_write_16 - write the 16 bit data at given io address */
static void io_write_16 (u16 io_port, u16 data)
{
    DP_OUTW (data, io_port);
}
#endif

/* io_read_32 - read the 32 bit value stored at given io address*/
static u32 io_read_32 (u16 io_port)
{
    return DP_INL(io_port);
}

/* io_write_32 - write the 32 bit data at given io address */
static void io_write_32 (u16 io_port, u32 data)
{
    DP_OUTL(data, io_port);
}

#if (DRV_DEBUG > 0)
/* dp83815_desc_info - print info on descriptors
 *
 * This routine displays values of the elements of dp83815_queue;
 * if option is -1, prints info on all the descs and their field vals;
 * and in all other cases prints info on the  specified descriptor.
 */

static void dp83815_desc_info (struct dp83815_queue * q, int option)
{
    virt_addr    desc_addr;
    u32          i;

    /* Display info header */
    printk ( KERN_DEBUG "dp83815_desc_info(): q=0x%x, "
             "first_desc_addr=0x%x, last_desc_addr=0x%x cur_desc_addr=0x%x\n",
             (u32)q,  (u32)q->first_desc_addr,
             (u32)q->last_desc_addr, (u32)q->cur_desc_addr);

    /* Print requested element info */
    switch (option) {
    case -1:                                 /* all elements */
        desc_addr = q->first_desc_addr;
        for (i=0; i<q->count; i++) {
            printk (KERN_DEBUG "desc_addr=0x%x link=0x%x "
                    "cmdsts=0x%x bufptr=0x%x\n",
                    (u32)desc_addr, DP_DESC_LNK_GET (desc_addr),
                    DP_DESC_CMDSTS_GET (desc_addr),
                    DP_DESC_BUFPTR_GET (desc_addr));
            desc_addr += DP_QUEUE_ELE_SIZE;
        }
        break;
        
    default:                                 /* a specific element */
        if (option > q->count) {
            printk (KERN_DEBUG "Can't print info for ele=%d when count=%d\n",
                    option, q->count);
        }
        else {
            desc_addr = q->first_desc_addr + (DP_QUEUE_ELE_SIZE * (option - 1));
            printk (KERN_DEBUG "ele=%d desc_addr=0x%x link=0x%x "
                    "cmdsts=0x%x bufptr=0x%x\n",
                    option, (u32)desc_addr, DP_DESC_LNK_GET (desc_addr),
                    DP_DESC_CMDSTS_GET (desc_addr),
                    DP_DESC_BUFPTR_GET (desc_addr));
        }
    }
}


/* dp83815_regs_info - prints values of registers */
static void dp83815_regs_info (u16 iobase)
{
    printk (KERN_INFO "dp(0x%x): CFG=0x%x IMR=0x%x IER=0x%x\n",
            iobase, DP_REG32_READ (DP_CFG), DP_REG32_READ (DP_IMR),
            DP_REG32_READ (DP_IER));
    printk (KERN_INFO " ++ TXDP=0x%x TXCFG=0x%x RXDP=0x%x RXCFG=0x%x\n",
            DP_REG32_READ (DP_TXDP), DP_REG32_READ (DP_TXCFG),
            DP_REG32_READ (DP_RXDP), DP_REG32_READ (DP_RXCFG));
}
#endif /* (DRV_DEBUG > 0) */

#ifdef MODULE

/*
 * init_module - Initialize the driver module
 *
 * This routine is called to initialize the driver when the module is
 * installed into a running kernel via insmod.
 *
 */

int
init_module(void)
{
    printk (KERN_INFO "%s", version);
    return dp83815_probe (NULL);
}

/*
 * cleanup_module - Deinstall the driver module.
 *
 * This routine is called when removing the driver module from a running
 * kernel via rmmod.
 *
 */

void
cleanup_module(void)
{
    struct device *cur_dev;

#ifdef DP_WOL_SUPPORT
    unregister_reboot_notifier(&dp83815_notifier);
    dp83815_shutdown_notifier(NULL, SYS_HALT, NULL); /* call it explicitly */
#endif

    cur_dev=dp83815_dev_list;
    while (cur_dev) {
        dp83815_dev_list = ((struct dp83815_priv *)(cur_dev->priv))->next;
        unregister_netdev (cur_dev);
        release_region (cur_dev->base_addr, DP83815_PCI_IO_SIZE);
        kfree (cur_dev);
        cur_dev = dp83815_dev_list;
    }
#ifdef CONFIG_COBALT_GEN_V
    lcd_handler_del(&dp83815_lcd_handler);
#endif
}
#endif /* MODULE */

#ifdef CONFIG_COBALT_GEN_V
static void dp83815_check_link(const void)
{
  struct device *cur_dev = dp83815_dev_list;

  while (cur_dev) {
    u32  iobase = cur_dev->base_addr;
    u16  phy_status = DP_REG16_READ (DP_PHYSTS);

    LED_SET_LINK(cur_dev, phy_status & DP_PHYSTS_LNK_VALID);
    cur_dev = ((struct dp83815_priv *)(cur_dev->priv))->next;
  }
}
#endif
